import asyncio
from big_thing_py.core.service import *
from big_thing_py.core.request import *
from big_thing_py.core.service_model import SkillValue
from big_thing_py.core.service_model import Objects as Skill


class MXValue(MXService):

    def __init__(
        self,
        func: Callable,
        tag_list: List[MXTag],
        type: MXType = None,
        category: SkillValue = Skill.testSkill.Values.testSkillValue,
        cycle: float = 0,
        bound: Tuple[float, float] = (),
        name: str = '',
        energy: float = 0,
        desc: str = '',
        thing_name: str = '',
        middleware_name: str = '',
        format: str = '',
    ) -> None:
        super().__init__(
            func=func,
            tag_list=tag_list,
            name=name,
            category=category,
            energy=energy,
            desc=desc,
            thing_name=thing_name,
            middleware_name=middleware_name,
        )
        self._type = type

        if self._type is None:
            # TODO (thsvkd): Not supported (ENUM, LIST, DICT) MXType yet
            if self._category.value_type.type in [MXType.ENUM, MXType.LIST, MXType.DICT]:
                self._type = MXType.STRING
            else:
                self._type = self._category.value_type.type
        elif self._type in [MXType.STRING, MXType.BINARY]:
            self._min, self._max = -1, -1

        if not bound:
            self._min, self._max = self._category.value_type.bound
        else:
            self._min, self._max = bound

        self._cycle = cycle  # if cycle is 0, it means that the value is event-based
        self._format = format

        self._last_value: Union[float, str, bool] = None
        self._last_update_time: float = 0
        self._binary_sending: bool = False

        self.is_initialized = False

        if len(get_function_info(self._func)['args']) > 0 or get_function_info(self._func)['return_type'] == None:
            raise MXValueError('callback function must not have any argument and must return value')
        if self._type in [MXType.UNDEFINED, MXType.VOID] or isinstance(self._type, str):
            raise MXValueError('type cannot be UNDEFINED or VOID or `str` type')
        if not self._type in [MXType.STRING, MXType.BINARY] and self._min >= self._max:
            raise MXValueError('bound must be min < max')
        if (self._cycle < 0) if self._cycle != None else False:
            raise MXValueError('cycle must be > 0 or same as 0 (event-based)')
        if not isinstance(self._format, str):
            raise MXValueError('format must be str')

        # self._category = MXValueCategory.get(self._func.__name__)

    def __eq__(self, o: 'MXValue') -> bool:
        instance_check = isinstance(o, MXValue)
        type_check = o._type == self._type
        bound_check = o._max == self._max and o._min == self._min
        format_check = o._format == self._format
        cycle_check = o._cycle == self._cycle

        return super().__eq__(o) and instance_check and type_check and bound_check and format_check and cycle_check

    def __getstate__(self):
        state = super().__getstate__()

        state['_type'] = self._type
        state['_min'] = self._min
        state['_max'] = self._max
        state['_cycle'] = self._cycle
        state['_format'] = self._format

        del state['_last_value']
        del state['_last_update_time']
        del state['_binary_sending']

        return state

    def __setstate__(self, state):
        super().__setstate__(state)

        self._type = state['_type']
        self._min = state['_min']
        self._max = state['_max']
        self._cycle = state['_cycle']
        self._format = state['_format']

        self._last_value = None
        self._last_update_time = 0
        self._binary_sending = False

    async def _async_wrapper(self):
        return self._func()

    async def update(self, set_value: Union[int, float, str, bool] = None) -> Union[int, float, str, bool]:
        try:
            if not set_value is None:
                new_value = set_value
            else:
                if asyncio.iscoroutinefunction(self._func):
                    new_value = await self._func()
                else:
                    new_value = await self._async_wrapper()

            if new_value is None:
                raise MXValueError('return value of Value callback function is None')

            self._last_update_time = get_current_datetime()

            if self._last_value != new_value:
                self._last_value = new_value
                return new_value
            else:
                return None
        except Exception as e:
            print_error(e)
        finally:
            if not self.is_initialized:
                self.is_initialized = True

    def dict(self) -> dict:
        return {
            'name': self._name,
            'category': self._category.value_id,
            'description': self._desc,
            'tags': [tag.dict() for tag in self._tag_list],
            'type': self._type.value,
            'bound': {'min_value': self._min, 'max_value': self._max},
            'format': self._format,
        }

    def publish_dict(self) -> dict:
        if isinstance(self._last_value, BinaryBase64):
            return {'type': self._type.value, 'value': str(self._last_value)}
        else:
            return {'type': self._type.value, 'value': self._last_value}

    def generate_value_publish_message(self) -> MXValuePublishMessage:
        return MXValuePublishMessage(self.name, self.thing_name, self.publish_dict())

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def type(self) -> MXType:
        return self._type

    @property
    def bound(self) -> Tuple[float, float]:
        return (self._min, self._max)

    @property
    def max(self) -> float:
        return self._max

    @property
    def min(self) -> float:
        return self._min

    @property
    def cycle(self) -> float:
        return self._cycle

    @property
    def format(self) -> str:
        return self._format

    @property
    def last_value(self) -> float:
        return self._last_value

    @property
    def last_update_time(self) -> float:
        return self._last_update_time

    @property
    def binary_sending(self) -> bool:
        return self._binary_sending

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @type.setter
    def type(self, type: MXType) -> None:
        self._type = type

    @bound.setter
    def bound(self, bound: Tuple[float, float]) -> None:
        if self._type in [MXType.STRING, MXType.BINARY]:
            self._min, self._max = -1, -1
        else:
            self._min = bound[0]
            self._max = bound[1]

    @max.setter
    def max(self, max: float) -> None:
        self._max = max

    @min.setter
    def min(self, min: float) -> None:
        self._min = min

    @cycle.setter
    def cycle(self, cycle: float) -> None:
        self._cycle = cycle

    @format.setter
    def format(self, format: str) -> None:
        self._format = format

    @last_value.setter
    def last_value(self, last_value: Union[float, str, bool]) -> None:
        self._last_value = last_value

    @last_update_time.setter
    def last_update_time(self, last_update_time: float) -> None:
        self._last_update_time = last_update_time

    @binary_sending.setter
    def binary_sending(self, binary_send: bool) -> bool:
        self._binary_sending = binary_send



from big_thing_py.common import *
from big_thing_py.core.request import *
from big_thing_py.core.mqtt_message import *
from big_thing_py.core.argument import *
from big_thing_py.core.service import *
from big_thing_py.core.service_model import SkillFunction, SkillFunctionArgument
from big_thing_py.core.service_model import Objects as Skill

import asyncio


class MXFunction(MXService):

    def __init__(
        self,
        func: Callable,
        tag_list: List[MXTag],
        return_type: MXType = None,
        category: Union[SkillValue, SkillFunction] = Skill.testSkill.Functions.testSkillFunction,
        name: str = '',
        energy: float = 0,
        desc: str = '',
        thing_name: str = '',
        middleware_name: str = '',
        arg_list: List[MXArgument] = None,
        exec_time: float = 0,
        timeout: float = 0,
        range_type: MXRangeType = MXRangeType.SINGLE,
    ) -> None:
        super().__init__(
            func=func,
            tag_list=tag_list,
            name=name,
            category=category,
            energy=energy,
            desc=desc,
            thing_name=thing_name,
            middleware_name=middleware_name,
        )

        # NOTE (thsvkd): generate getter MXFunction instance of target MXValue
        if issubclass(self._category, SkillValue):
            self._return_type = self._category.value_type.type
            self._arg_list = []
        else:
            self._arg_list = arg_list
            if self._arg_list is None:
                arguments: List[SkillFunctionArgument] = self._category.arguments
                self._arg_list = [
                    MXArgument(
                        name=arg.argument_id,
                        type=arg.argument_type.type,
                        bound=arg.argument_type.bound,
                    )
                    for arg in arguments
                ]

        # TODO (thsvkd): Not supported (ENUM, LIST, DICT) MXType yet
        if not return_type is None:
            self._return_type = return_type
        elif self._category.return_type in [MXType.ENUM, MXType.LIST, MXType.DICT]:
            self._return_type = MXType.STRING
        else:
            self._return_type = self._category.return_type

        self._exec_time = exec_time
        self._timeout = timeout
        # TODO: range_type will be removed from MXFunction
        self._range_type = range_type

        self._return_value = None
        self._running = False
        self._running_scenario_list: List[str] = []

        if self._return_type in [MXType.UNDEFINED] or isinstance(self._return_type, str):
            raise MXValueError('return_type cannot be undefined or `str` type')
        if (not len(self._arg_list) == len(get_function_info(self._func)['args'])) if self._func else False:
            raise MXValueError(
                f'Length of argument list must be same with callback function.\n'
                f'given: {len(self._arg_list)}, expected: {len(get_function_info(self._func)["args"])}'
            )

        # TODO: return type detection feature
        # if not self._return_type or self._return_type in [MXType.UNDEFINED]:
        #     func_info = get_function_info(self._func)
        #     self._return_type = MXType().get(func_info[2])
        #     raise MXValueError('type must be specified')
        # elif self._return_type in [MXType.UNDEFINED]:
        #     raise MXValueError('type cannot be undefined or void')

    def __eq__(self, o: 'MXFunction') -> bool:
        instance_check = isinstance(o, MXFunction)
        arg_list_check = o._arg_list == self._arg_list
        return_type_check = o._return_type == self._return_type
        exec_time_check = o._exec_time == self._exec_time
        timeout_check = o._timeout == self._timeout
        range_type_check = o._range_type == self._range_type

        return (
            super().__eq__(o) and instance_check and arg_list_check and return_type_check and exec_time_check and timeout_check and range_type_check
        )

    def __getstate__(self):
        state = super().__getstate__()

        state['_return_type'] = self._return_type
        state['_arg_list'] = self._arg_list
        state['_exec_time'] = self._exec_time
        state['_timeout'] = self._timeout
        state['_range_type'] = self._range_type

        del state['_return_value']
        del state['_running']
        del state['_running_scenario_list']

        return state

    def __setstate__(self, state):
        super().__setstate__(state)

        self._type = state['_type']
        self._min = state['_min']
        self._max = state['_max']
        self._cycle = state['_cycle']
        self._format = state['_format']

        self._return_value = None
        self._running = False
        self._running_scenario_list = []

    async def _async_wrapper(self, *args, **kwargs):
        return self._func(*args, **kwargs)

    async def execute(self, execute_request: Union[MXExecuteRequest, MXInnerExecuteRequest]) -> Tuple['MXFunction', MXErrorCode]:
        execute_msg = execute_request.trigger_msg
        self._running_scenario_list.append(execute_msg.scenario)
        execute_request.timer_start()

        execute_msg = execute_request._trigger_msg
        MXLOG_DEBUG(f'[FUNC RUN] run {self._name} function by {execute_msg.scenario}', 'green')

        try:
            if not isinstance(execute_msg, MXExecuteMessage):
                raise Exception(f'[{get_current_function_name()}] Wrong ExecuteMessage type - execute_msg: {type(execute_msg)}')

            self._running = True
            execute_request.result_msg.error = MXErrorCode.NO_ERROR

            if self._timeout:
                if asyncio.iscoroutinefunction(self._func):
                    self._return_value = await asyncio.wait_for(self._func(*execute_msg.tuple_arguments()), timeout=self._timeout)
                else:
                    self._return_value = await asyncio.wait_for(self._async_wrapper(*execute_msg.tuple_arguments()), timeout=self._timeout)
            else:
                if asyncio.iscoroutinefunction(self._func):
                    self._return_value = await self._func(*execute_msg.tuple_arguments())
                else:
                    self._return_value = await self._async_wrapper(*execute_msg.tuple_arguments())
        except KeyboardInterrupt as e:
            # TODO: for wrapup main thread, but not test it yet
            print_error(e)
            MXLOG_DEBUG('Function execution exit by user', 'red')
            raise e
        except asyncio.CancelledError as e:
            # TODO: for wrapup main thread, but not test it yet
            print_error(e)
            MXLOG_DEBUG('Function execution exit by user', 'red')
            raise e
        except asyncio.TimeoutError as e:
            MXLOG_DEBUG(f'[FUNC TIMEOUT] function {self._name} by scenario {execute_msg.scenario} was timeout!!!', 'yellow')
            execute_request.result_msg.error = MXErrorCode.TIMEOUT
        except BaseException as e:
            print_error(e)
            MXLOG_DEBUG(
                f'[FUNC FAIL] function {self._name} by scenario {execute_msg.scenario} is failed while executing!!!',
                'red',
            )
            execute_request.result_msg.error = MXErrorCode.FAIL
        else:
            execute_request.result_msg.error = MXErrorCode.NO_ERROR
        finally:
            if isinstance(self._return_value, str) and is_base64(self._return_value) and self._return_type != MXType.BINARY:
                raise MXValueError('return value is base64, but type is not BINARY')

            self._running = False
            self._running_scenario_list.remove(execute_msg.scenario)

            execute_request.result_msg.return_value = self._return_value
            execute_request.timer_end()
            MXLOG_DEBUG(
                f'[FUNC END] function {self._name} end. -> return value : {self._return_value}, duration: {execute_request.duration():.4f} Sec',
                'green',
            )

            return execute_request

    def dict(self) -> dict:
        return {
            'name': self._name,
            'category': self._category.function_id if issubclass(self._category, SkillFunction) else self._category.value_id,
            'description': self._desc,
            'exec_time': self._exec_time * 1000 if self._exec_time is not None else 0,
            'return_type': self._return_type.value,
            'energy': self._energy,
            'tags': [tag.dict() for tag in self._tag_list],
            'use_arg': 1 if len(self._arg_list) > 0 else 0,
            'arguments': [argument.dict() for argument in self._arg_list] if self._arg_list else [],
        }

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    def generate_execute_result_message(
        self,
        execute_request: Union[MXExecuteRequest, MXInnerExecuteRequest],
    ) -> MXExecuteResultMessage:
        trigger_msg = execute_request.trigger_msg
        result_msg = execute_request.result_msg

        if isinstance(result_msg.return_value, BinaryBase64):
            return_value = str(result_msg.return_value)
        else:
            return_value = result_msg.return_value

        return MXExecuteResultMessage(
            function_name=trigger_msg.function_name,
            thing_name=result_msg.thing_name,
            middleware_name=result_msg.middleware_name,
            scenario=trigger_msg.scenario,
            client_id=trigger_msg.client_id,
            request_ID=trigger_msg.request_ID,
            return_type=result_msg.return_type,
            return_value=return_value,
            error=result_msg.error,
            action_type=execute_request.action_type,
        )

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def exec_time(self) -> float:
        return self._exec_time

    @property
    def timeout(self) -> float:
        return self._timeout

    @property
    def arg_list(self) -> List[MXArgument]:
        return self._arg_list

    @property
    def return_type(self) -> MXType:
        return self._return_type

    @property
    def return_value(self) -> Union[int, float, str, bool]:
        return self._return_value

    @property
    def running(self) -> bool:
        return self._running

    @property
    def range_type(self) -> MXRangeType:
        return self._range_type

    @property
    def running_scenario_list(self) -> List[str]:
        return self._running_scenario_list

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @exec_time.setter
    def exec_time(self, exec_time: float) -> None:
        self._exec_time = exec_time

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        self._timeout = timeout

    @arg_list.setter
    def arg_list(self, arg_list: List[MXArgument]) -> None:
        self._arg_list = arg_list

    @return_type.setter
    def return_type(self, return_type: MXType) -> None:
        self._return_type = return_type

    @return_value.setter
    def return_value(self, return_value: Union[int, float, str, bool]) -> None:
        self._return_value = return_value

    @running.setter
    def running(self, running: bool) -> None:
        self._running = running

    @range_type.setter
    def range_type(self, range_type: MXRangeType) -> None:
        self._range_type = range_type


if __name__ == '__main__':
    pass



from big_thing_py.utils import *


class MXArgument:
    def __init__(self, name: str, type: MXType, bound: Tuple[float, float] = ()):
        self._name = name
        self._type = type

        # TODO (thsvkd): Not supported (ENUM, LIST, DICT) MXType yet
        if self._type in [MXType.ENUM, MXType.LIST, MXType.DICT]:
            self._type = MXType.STRING

        if self._type in [MXType.STRING, MXType.BINARY]:
            self._min, self._max = -1, -1
        else:
            self._min, self._max = bound
            if self._min >= self._max:
                raise MXValueError('bound must be min < max')

        if check_valid_identifier(self._name) == MXErrorCode.INVALID_DATA:
            raise MXValueError(f'name cannot be empty & can only contain alphanumeric characters and underscores. name: {self._name}')
        elif check_valid_identifier(self._name) == MXErrorCode.TOO_LONG_IDENTIFIER:
            raise MXValueError(f'too long identifier. name: {self._name}, length: {len(self._name)}')

        if self._type in [MXType.UNDEFINED, MXType.VOID] or isinstance(self._type, str):
            raise MXValueError('type cannot be UNDEFINED or VOID or `str` type')

    def __str__(self) -> str:
        return self._name

    def __eq__(self, o: 'MXArgument') -> bool:
        instance_check = isinstance(o, MXArgument)
        name_check = o._name == self._name
        type_check = o._type == self._type
        min_check = o._min == self._min
        max_check = o._max == self._max

        return instance_check and name_check and type_check and min_check and max_check

    def dict(self) -> dict:
        return {'name': self._name, 'type': self._type.value, 'bound': {'min_value': self._min, 'max_value': self._max}}

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> MXType:
        return self._type

    @property
    def bound(self) -> Tuple[float, float]:
        return self._min, self._max

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @name.setter
    def name(self, name: str):
        self._name = name

    @type.setter
    def type(self, type: MXType):
        self._type = type

    @bound.setter
    def bound(self, bound: Tuple[float, float]):
        if self._type in [MXType.STRING, MXType.BINARY]:
            self._min, self._max = -1, -1
        else:
            self._min = bound[0]
            self._max = bound[1]



from aenum import IntEnum, StrEnum


class MXStrEnum(StrEnum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class MXIntEnum(IntEnum):
    def _generate_next_value_(name, start, count, last_values):
        return count



from big_thing_py.utils import *


def make_super_request_key(scenario_name: str, requester_middleware_name: str) -> str:
    return '@'.join([scenario_name, requester_middleware_name])


def make_sub_service_request_key(sub_service_name: str, sub_service_request_order: int) -> str:
    return '@'.join([sub_service_name, str(sub_service_request_order)])


def make_request_ID(requester_middleware_name: str, super_thing_name: str, super_service_name: str, sub_service_request_order: int):
    return '@'.join([requester_middleware_name, super_thing_name, super_service_name, str(sub_service_request_order)])


class MXMQTTMessage:
    def __init__(self, protocol_type: Union[MXProtocolType.Base, MXProtocolType.Super, MXProtocolType.WebClient]) -> None:
        self.protocol_type = protocol_type
        self.topic: str = None
        self.timestamp: float = None

        self.topic_error: bool = False
        self.payload_error: bool = False

    def set_timestamp(self, timestamp: float = None) -> None:
        if timestamp:
            self.timestamp = timestamp
        else:
            self.timestamp = time.time()


class MXMQTTSendMessage(MXMQTTMessage):
    def __init__(self, topic: str, payload: dict, protocol_type: Union[MXProtocolType.Base, MXProtocolType.Super, MXProtocolType.WebClient]) -> None:
        super().__init__(protocol_type)
        self.topic = topic
        self.payload = payload

        if MXProtocolType.get(self.topic) != self.protocol_type:
            self.topic_error = True
        else:
            self.topic_error = False

    def mqtt_message(self) -> MQTTMessage:
        msg = encode_MQTT_message(self.topic, self.payload)
        return msg


class MXMQTTReceiveMessage(MXMQTTMessage):
    def __init__(self, msg: MQTTMessage, protocol_type: Union[MXProtocolType.Base, MXProtocolType.Super, MXProtocolType.WebClient]) -> None:
        super().__init__(protocol_type)
        self.topic, self.payload = decode_MQTT_message(msg)

        if MXProtocolType.get(self.topic) != self.protocol_type:
            self.topic_error = True
        else:
            self.topic_error = False


##############################################################################################################################


class MXRegisterMessage(MXMQTTSendMessage):
    def __init__(self, thing_name: str, payload: dict) -> None:
        protocol_type = MXProtocolType.Base.TM_REGISTER
        topic = protocol_type.value % (thing_name)
        super().__init__(topic=topic, payload=payload, protocol_type=protocol_type)


class MXRequestRegisterInfoMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg, protocol_type=MXProtocolType.Base.MT_REQUEST_REGISTER_INFO)
        self.middleware_name = self.topic.split('/')[2]
        self.error: MXErrorCode = None

        self.client_id: str = self.payload.get('client_id', None)
        if self.client_id is None:
            self.payload_error = True


class MXRequestUnregisterMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg, protocol_type=MXProtocolType.Base.MT_REQUEST_UNREGISTER)
        self.middleware_name = self.topic.split('/')[2]
        self.error: MXErrorCode = None

        self.client_id: str = self.payload.get('client_id', None)
        if self.client_id is None:
            self.payload_error = True


class MXRegisterResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg, protocol_type=MXProtocolType.Base.MT_RESULT_REGISTER)
        self.thing_name = self.topic.split('/')[3]
        self.middleware_name: str = self.payload['middleware_name']
        self.error: MXErrorCode = MXErrorCode.get(self.payload['error'])


class MXUnregisterMessage(MXMQTTSendMessage):
    def __init__(self, thing_name: str) -> None:
        protocol_type = MXProtocolType.Base.TM_UNREGISTER
        topic = protocol_type.value % (thing_name)
        payload = EMPTY_JSON
        super().__init__(topic=topic, payload=payload, protocol_type=protocol_type)


class MXUnregisterResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg, protocol_type=MXProtocolType.Base.MT_RESULT_UNREGISTER)
        self.thing_name: str = self.topic.split('/')[3]
        self.error: MXErrorCode = MXErrorCode.get(self.payload['error'])


class MXExecuteMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg, protocol_type=MXProtocolType.get(msg.topic.decode()))

        if self.protocol_type in [MXProtocolType.Base.MT_EXECUTE, MXProtocolType.Base.MT_IN_EXECUTE]:
            self.topic_error = False
        else:
            self.topic_error = True
            return

        self.middleware_name: str = ''
        self.request_ID: str = ''
        self.client_id: str = ''

        if self.protocol_type == MXProtocolType.Base.MT_EXECUTE:
            self.function_name: str = self.topic.split('/')[2]
            self.thing_name: str = self.topic.split('/')[3]

            if len(self.topic.split('/')) == 6:
                # MT/EXECUTE/[FunctionName]/[ThingName]/[TargetMiddlewareName]/[Request_ID] topic
                self.middleware_name: str = self.topic.split('/')[4]
                self.request_ID: str = self.topic.split('/')[5]
                if not self.middleware_name or not self.request_ID:
                    self.topic_error = True
            elif len(self.topic.split('/')) == 4:
                # MT/EXECUTE/[FunctionName]/[ThingName] topic
                self.middleware_name: str = ''
                self.request_ID: str = ''
            else:
                self.topic_error = True
        elif self.protocol_type == MXProtocolType.Base.MT_IN_EXECUTE:
            self.function_name: str = self.topic.split('/')[3]
            self.thing_name: str = self.topic.split('/')[4]
            self.middleware_name: str = ''
            self.request_ID: str = ''

        self.scenario: str = self.payload.get('scenario', '')
        self.arguments: List[dict] = self.payload.get('arguments', [])
        self.client_id: str = self.payload.get('client_id', '')

        if not self.scenario or self.arguments == None:
            self.payload_error = True
        elif not all([isinstance(arg.get('order', None), int) for arg in self.arguments]) or not all(
            [isinstance(arg.get('value', None), (int, float, str, bool)) for arg in self.arguments]
        ):
            self.payload_error = True

    def tuple_arguments(self) -> tuple:
        self.arguments = sorted(self.arguments, key=lambda x: int(x['order']))
        real_arguments = tuple([argument['value'] for argument in self.arguments])
        return real_arguments

    def dict_arguments(self) -> List[dict]:
        self.arguments = sorted(self.arguments, key=lambda x: int(x['order']))
        json_arguments = [dict(order=arg['order'], value=arg['value']) for arg in self.arguments]
        return json_arguments


class MXExecuteResultMessage(MXMQTTSendMessage):
    def __init__(
        self,
        function_name: str = '',
        thing_name: str = '',
        middleware_name: str = '',
        scenario: str = '',
        client_id: str = '',
        request_ID: str = '',
        return_type: MXType = MXType.UNDEFINED,
        return_value: Union[int, float, str, bool] = None,
        error: MXErrorCode = MXErrorCode.UNDEFINED,
        action_type: MXActionType = MXActionType.EXECUTE,
    ) -> None:
        if action_type == MXActionType.EXECUTE:
            protocol_type = MXProtocolType.Base.TM_RESULT_EXECUTE
        elif action_type == MXActionType.INNER_EXECUTE:
            protocol_type = MXProtocolType.Base.TM_IN_RESULT_EXECUTE
        else:
            self.topic_error = True
            return

        self.function_name = function_name
        self.thing_name = thing_name
        self.middleware_name = middleware_name
        self.scenario = scenario
        self.client_id = client_id
        self.request_ID = request_ID
        self.return_type = return_type
        self.return_value = return_value
        self._error = error

        if self.request_ID:
            topic = protocol_type.value % (
                self.function_name,
                self.thing_name,
                self.middleware_name,
                self.request_ID,
            )
        else:
            if action_type == MXActionType.EXECUTE:
                topic = (protocol_type.value % (self.function_name, self.thing_name, '', '')).rstrip('/')
            elif action_type == MXActionType.INNER_EXECUTE:
                topic = protocol_type.value % (self.function_name, self.thing_name)

        if action_type == MXActionType.EXECUTE:
            payload = dict(
                error=self._error.value,
                scenario=self.scenario,
                return_type=self.return_type.value,
                return_value=self.return_value,
            )
        elif action_type == MXActionType.INNER_EXECUTE:
            payload = dict(
                error=self._error.value,
                scenario=self.scenario,
                client_id=self.client_id,
                return_type=self.return_type.value,
                return_value=self.return_value,
            )

        super().__init__(topic=topic, payload=payload, protocol_type=protocol_type)

    @property
    def error(self) -> dict:
        return self._error

    @error.setter
    def error(self, error: MXErrorCode):
        if not isinstance(error, MXErrorCode):
            raise MXTypeError("error must be an MXErrorType")
        self._error = error
        self.payload = dict(
            error=self._error.value,
            scenario=self.scenario,
            return_type=self.return_type.value,
            return_value=self.return_value,
        )


# TODO: binary 부분 구현 완료 후 작성하기
class MXBinaryValueResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg, protocol_type=MXProtocolType.Base.MT_RESULT_BINARY_VALUE)
        self.thing_name = self.topic.split('/')[3]
        self.value_name = self.payload['value']


class MXAliveMessage(MXMQTTSendMessage):
    def __init__(self, thing_name: str) -> None:
        protocol_type = MXProtocolType.Base.TM_ALIVE
        topic = protocol_type.value % (thing_name)
        payload = EMPTY_JSON
        super().__init__(topic=topic, payload=payload, protocol_type=protocol_type)


class MXValuePublishMessage(MXMQTTSendMessage):
    def __init__(self, value_name: str, thing_name: str, payload: dict) -> None:
        protocol_type = MXProtocolType.Base.TM_VALUE_PUBLISH
        topic = protocol_type.value % (value_name, thing_name)
        super().__init__(topic=topic, payload=payload, protocol_type=protocol_type)


# for middleware detect
class MXHomeResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg, protocol_type=MXProtocolType.WebClient.ME_RESULT_HOME)
        self._client_id = self.topic.split('/')[3]


class MXGetHomeMessage(MXMQTTSendMessage):
    def __init__(self, thing_name: str) -> None:
        protocol_type = MXProtocolType.WebClient.EM_GET_HOME
        topic = protocol_type.value % (thing_name)
        payload = EMPTY_JSON
        super().__init__(topic=topic, payload=payload, protocol_type=protocol_type)


class MXNotifyMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg, protocol_type=MXProtocolType.WebClient.ME_NOTIFY_CHANGE)

        # topic
        self._client_id = self.topic.split('/')[2]

        # payload


if __name__ == '__main__':
    payload = {
        'scenario': 'test',
        'arguments': [
            {
                'order': 0,
                'value': 1,
            },
            {
                'order': 1,
                'value': 2,
            },
        ],
    }
    msg = encode_MQTT_message('MT/EXECUTE/test1/test2', payload)
    message = MXExecuteMessage(msg)
    print(message.tuple_arguments())
    print(message.dict_arguments(*(1, 2)))



from big_thing_py.big_thing import *
from big_thing_py.super import *
import threading


class MXSuperThing(MXBigThing):
    DEFAULT_NAME = 'default_super_thing'

    # Super Service Execution 요청이 들어왔을때 mapping_table에 있는 super_request를 찾기 위한 super_service_request_key 리스트
    # Super Service는 자신의 이름으로 super_request_key를 찾을 수 있다.
    # {
    #     'super_service_request_key1': ['sub_service_request_key1', 'sub_service_request_key2', ...]},
    #     'super_service_request_key2': ['sub_service_request_key3', 'sub_service_request_key4', ...]},
    #      ...
    # }
    _SUPER_SERVICE_REQUEST_KEY_TABLE: Dict[str, List[str]] = dict()

    def __init__(
        self,
        name: str = DEFAULT_NAME,
        desc: str = '',
        version: str = sdk_version(),
        service_list: List[MXService] = [],
        alive_cycle: float = 60,
        is_super: bool = True,
        is_parallel: bool = True,
        ip: str = '127.0.0.1',
        port: int = 1883,
        ssl_ca_path: str = '',
        ssl_cert_path: str = '',
        ssl_key_path: str = '',
        log_name: str = '',
        log_enable: bool = True,
        log_mode: MXPrintMode = MXPrintMode.ABBR,
        append_mac_address: bool = True,
        refresh_cycle: float = 30,
    ):
        self._global_service_table: Dict[str, Union[List[MXFunction], List[MXValue]]] = dict(values=[], functions=[])
        self._SUPER_SERVICE_REQUEST_KEY_TABLE = dict()

        super().__init__(
            name=name,
            desc=desc,
            version=version,
            service_list=service_list,
            alive_cycle=alive_cycle,
            is_super=is_super,
            is_parallel=is_parallel,
            ip=ip,
            port=port,
            ssl_ca_path=ssl_ca_path,
            ssl_cert_path=ssl_cert_path,
            ssl_key_path=ssl_key_path,
            log_name=log_name,
            log_enable=log_enable,
            log_mode=log_mode,
            append_mac_address=append_mac_address,
        )

        self._refresh_cycle = refresh_cycle

        self._last_refresh_time = 0

        self._task_func_list += [
            self._refresh_thread_func,
        ]

    def __eq__(self, o: 'MXSuperThing'):
        instance_check = isinstance(o, MXSuperThing)
        refresh_cycle_check = self._refresh_cycle == o._refresh_cycle

        return super().__eq__(o) and instance_check and refresh_cycle_check

    def __getstate__(self):
        state = super().__getstate__()

        state['_refresh_cycle'] = self._refresh_cycle

        del state['_last_refresh_time']

        return state

    def __setstate__(self, state):
        super().__setstate__(state)

        self._refresh_cycle = state['_refresh_cycle']

        self._last_refresh_time = 0
        self._task_func_list = self._task_func_list + [
            self._refresh_thread_func,
        ]

    @override
    def _setup(self) -> 'MXSuperThing':
        self._init_logger()
        self._extract_sub_service_request_info()
        return super()._setup()

    # ===========================================================================================
    #  _    _                             _    __                      _    _
    # | |  | |                           | |  / _|                    | |  (_)
    # | |_ | |__   _ __   ___   __ _   __| | | |_  _   _  _ __    ___ | |_  _   ___   _ __   ___
    # | __|| '_ \ | '__| / _ \ / _` | / _` | |  _|| | | || '_ \  / __|| __|| | / _ \ | '_ \ / __|
    # | |_ | | | || |   |  __/| (_| || (_| | | |  | |_| || | | || (__ | |_ | || (_) || | | |\__ \
    #  \__||_| |_||_|    \___| \__,_| \__,_| |_|   \__,_||_| |_| \___| \__||_| \___/ |_| |_||___/
    # ===========================================================================================

    def _refresh_thread_func(self, stop_event: Event):
        try:
            while not stop_event.wait(THREAD_TIME_OUT):
                if self._is_registered.is_set():
                    if (get_current_datetime() - self._last_refresh_time) > self._refresh_cycle:
                        self._send_SM_REFRESH()
                        self._last_refresh_time = get_current_datetime()
                        # time.sleep(self._refresh_cycle / 2)

        except Exception as e:
            stop_event.set()
            print_error(e)
            return False

        except Exception as e:
            stop_event.set()
            print_error(e)
            return False

        # def _unsubscribe_thread_func(self, stop_event: Event):
        #     try:
        #         while not stop_event.wait(THREAD_TIME_OUT):
        #             if self._is_registered:
        #                 if not self._unsubscribe_queue.empty():
        #                     topic = self._unsubscribe_queue.queue[0]
        #                     self._unsubscribe(topic)
        #                     topic = self._unsubscribe_queue.get()

        except Exception as e:
            stop_event.set()
            print_error(e)
            return False

    # ======================================================================================================================= #
    #  _    _                    _  _         __  __   ____  _______  _______   __  __                                        #
    # | |  | |                  | || |       |  \/  | / __ \|__   __||__   __| |  \/  |                                       #
    # | |__| |  __ _  _ __    __| || |  ___  | \  / || |  | |  | |      | |    | \  / |  ___  ___  ___   __ _   __ _   ___    #
    # |  __  | / _` || '_ \  / _` || | / _ \ | |\/| || |  | |  | |      | |    | |\/| | / _ \/ __|/ __| / _` | / _` | / _ \   #
    # | |  | || (_| || | | || (_| || ||  __/ | |  | || |__| |  | |      | |    | |  | ||  __/\__ \\__ \| (_| || (_| ||  __/   #
    # |_|  |_| \__,_||_| |_| \__,_||_| \___| |_|  |_| \___\_\  |_|      |_|    |_|  |_| \___||___/|___/ \__,_| \__, | \___|   #
    #                                                                                                         __/ |           #
    #                                                                                                         |___/           #
    # ======================================================================================================================= #

    @override
    def _handle_mqtt_message(self, msg: MQTTMessage) -> bool:
        topic_string = decode_MQTT_message(msg)[0]
        protocol = MXProtocolType.get(topic_string)

        if protocol == MXProtocolType.Super.MS_RESULT_SCHEDULE:
            rc = self._handle_MS_RESULT_SCHEDULE(msg)
        elif protocol == MXProtocolType.Super.MS_RESULT_EXECUTE:
            rc = self._handle_MS_RESULT_EXECUTE(msg)
        elif protocol == MXProtocolType.Super.MS_RESULT_SERVICE_LIST:
            rc = self._handle_MS_RESULT_SERVICE_LIST(msg)
        elif protocol == MXProtocolType.Super.MS_SCHEDULE:
            rc = self._handle_MS_SCHEDULE(msg)
        elif protocol == MXProtocolType.Super.MS_EXECUTE:
            rc = self._handle_MS_EXECUTE(msg)
        elif protocol == MXProtocolType.WebClient.ME_NOTIFY_CHANGE:
            rc = self._handle_ME_NOTIFY(msg)

        elif protocol == MXProtocolType.Base.MT_RESULT_REGISTER:
            rc = self._handle_MT_RESULT_REGISTER(msg)
        elif protocol == MXProtocolType.Base.MT_RESULT_UNREGISTER:
            rc = self._handle_MT_RESULT_UNREGISTER(msg, target_thing=self._thing_data)
        elif protocol == MXProtocolType.Base.MT_RESULT_BINARY_VALUE:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Not permitted topic! topic: {topic_string}')
            return False
        elif protocol == MXProtocolType.Base.MT_EXECUTE:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Not permitted topic! topic: {topic_string}')
            return False
        else:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Unexpected topic! topic: {topic_string}')
            return False

        if not rc:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Unexpected topic! topic: {topic_string}')
            return False

        return rc

    # ================
    # ___  ___ _____
    # |  \/  |/  ___|
    # | .  . |\ `--.
    # | |\/| | `--. \
    # | |  | |/\__/ /
    # \_|  |_/\____/
    # ================

    def _handle_MS_SCHEDULE(self, msg: MQTTMessage) -> bool:
        super_schedule_msg = MXSuperScheduleMessage(msg)
        target_super_service = self._get_function(super_schedule_msg.super_service_name)

        if not target_super_service:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Service {super_schedule_msg.super_service_name} does not exist...',
                'red',
            )
            return False
        if self._name != super_schedule_msg.super_thing_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Thing name {super_schedule_msg.super_thing_name} is not matched...',
                'red',
            )
            return False
        if self._middleware_name != super_schedule_msg.super_middleware_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Middleware name {super_schedule_msg.super_middleware_name} is not matched...',
                'red',
            )
            return False
        if super_schedule_msg.topic_error or super_schedule_msg.payload_error:
            MXLOG_DEBUG(f'[{get_current_function_name()}] super_schedule_msg Message has error!', 'red')
            return False
        if not self._check_super_service_exist(target_super_service):
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Service {target_super_service.name} does not exist...',
                'red',
            )
            return False

        schedule_thread = target_super_service.start_super_schedule_thread(super_schedule_msg, self._global_service_table, timeout=1000)
        if not schedule_thread:
            return False
        elif schedule_thread.is_alive():
            return True
        else:
            return False

    def _handle_MS_EXECUTE(self, msg: MQTTMessage) -> bool:
        super_execute_msg = MXSuperExecuteMessage(msg)
        target_super_service = self._get_function(super_execute_msg.super_service_name)

        if not target_super_service:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Service {super_execute_msg.super_service_name} does not exist...',
                'red',
            )
            return False
        if self._name != super_execute_msg.super_thing_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Thing name {super_execute_msg.super_thing_name} is not matched...',
                'red',
            )
            return False
        if self._middleware_name != super_execute_msg.super_middleware_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Middleware name {super_execute_msg.super_middleware_name} is not matched...',
                'red',
            )
            return False
        if super_execute_msg.topic_error or super_execute_msg.payload_error:
            MXLOG_DEBUG(f'[{get_current_function_name()}] super_execute_msg Message has error!', 'red')
            return False

        # 중복된 시나리오로부터 온 실행 요청이면 -4 에러코드를 보낸다.
        if super_execute_msg.scenario in target_super_service.running_scenario_list:
            target_super_service._send_SM_RESULT_EXECUTE(
                super_service_name=super_execute_msg.super_service_name,
                super_thing_name=super_execute_msg.super_thing_name,
                super_middleware_name=super_execute_msg.super_middleware_name,
                requester_middleware_name=super_execute_msg.requester_middleware_name,
                scenario=super_execute_msg.scenario,
                error=MXErrorCode.DUPLICATE,
            )
            return True

        super_execute_thread = target_super_service.start_super_execute_thread(super_execute_msg, self._SUPER_SERVICE_REQUEST_KEY_TABLE)
        if not super_execute_thread:
            return False
        else:
            return True

    def _handle_MS_RESULT_SCHEDULE(self, msg: MQTTMessage) -> bool:
        subschedule_result_msg = MXSubScheduleResultMessage(msg)
        subschedule_result_msg.set_timestamp()
        super_service_request_key = make_super_request_key(
            scenario_name=subschedule_result_msg.scenario,
            requester_middleware_name=subschedule_result_msg.requester_middleware_name,
        )
        sub_service_request_key = make_sub_service_request_key(
            sub_service_name=subschedule_result_msg.sub_service_name,
            sub_service_request_order=subschedule_result_msg.sub_service_request_order,
        )

        target_super_service = self._get_function(subschedule_result_msg.super_service_name)
        if not target_super_service:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Service {subschedule_result_msg.super_service_name} does not exist...',
                'yellow',
            )
            return False

        result = target_super_service.put_subschedule_result(super_service_request_key, sub_service_request_key, subschedule_result_msg)
        return result

    def _handle_MS_RESULT_EXECUTE(self, msg: MQTTMessage) -> bool:
        sub_service_execute_result_msg = MXSubExecuteResultMessage(msg)
        sub_service_execute_result_msg.set_timestamp()
        super_service_request_key = make_super_request_key(
            scenario_name=sub_service_execute_result_msg.scenario,
            requester_middleware_name=sub_service_execute_result_msg.requester_middleware_name,
        )
        sub_service_request_key = make_sub_service_request_key(
            sub_service_name=sub_service_execute_result_msg.sub_service_name,
            sub_service_request_order=sub_service_execute_result_msg.sub_service_request_order,
        )

        target_super_service = self._get_function(sub_service_execute_result_msg.super_service_name)
        if not target_super_service:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Super Service {sub_service_execute_result_msg.super_service_name} does not exist...',
                'yellow',
            )
            return False

        result = target_super_service.put_sub_service_execute_result(
            super_service_request_key, sub_service_request_key, sub_service_execute_result_msg
        )
        return result

    def _handle_MS_RESULT_SERVICE_LIST(self, msg: MQTTMessage) -> bool:
        try:
            service_list = MXSuperServiceListResultMessage(msg)
            service_list.set_timestamp()

            for middleware in service_list.service_list:
                hierarchy_type = middleware['hierarchy']
                if not hierarchy_type in ['local', 'child']:
                    MXLOG_DEBUG(f'[{get_current_function_name()}] Parent middleware is not supported', 'red')
                    return False

                middleware_name = middleware['middleware']
                if not middleware_name:
                    MXLOG_DEBUG(f'[{get_current_function_name()}] Middleware name does not exist', 'red')
                    return False

                thing_list = middleware['things']
                for thing in thing_list:
                    is_alive = thing['is_alive']
                    if is_alive != 1:
                        continue

                    is_super = thing['is_super']
                    alive_cycle = thing['alive_cycle']

                    # value 정보를 추출
                    value_service_list = self._extract_value_info(thing=thing, middleware_name=middleware_name)
                    self._global_service_table['values'].extend(value_service_list)

                    function_service_list = self._extract_function_info(thing_info=thing, middleware_name=middleware_name)
                    self._global_service_table['functions'].extend(function_service_list)

            self._last_refresh_time = get_current_datetime()
            return True
        except KeyError as e:
            print_error(e)
            MXLOG_DEBUG(f'[{get_current_function_name()}] KeyError', 'red')
            return False
        except ValueError as e:
            print_error(e)
            MXLOG_DEBUG(f'[{get_current_function_name()}] ValueError', 'red')
            return False
        except Exception as e:
            print_error(e)
            MXLOG_DEBUG(f'[{get_current_function_name()}] Unknown Exception', 'red')
            return False

    # ===================
    #   __  __   ______
    #  |  \/  | |  ____|
    #  | \  / | | |__
    #  | |\/| | |  __|
    #  | |  | | | |____
    #  |_|  |_| |______|
    # ===================

    def _handle_ME_NOTIFY(self, msg: MQTTMessage):
        notify_msg = MXNotifyMessage(msg)
        notify_msg.set_timestamp()
        self._send_SM_REFRESH()

    # ================
    #  _____ ___  ___
    # /  ___||  \/  |
    # \ `--. | .  . |
    #  `--. \| |\/| |
    # /\__/ /| |  | |
    # \____/ \_|  |_/
    # ================

    def _send_SM_EXECUTE(self, sub_service_execute_msg: MXSubExecuteMessage) -> None:
        sub_service_execute_mqtt_msg = sub_service_execute_msg.mqtt_message()
        self._publish_queue.put(sub_service_execute_mqtt_msg)

    def _send_SM_REFRESH(self):
        super_refresh_msg = self.generate_super_refresh_message()
        super_refresh_mqtt_msg = super_refresh_msg.mqtt_message()
        self._publish_queue.put(super_refresh_mqtt_msg)

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    def generate_super_refresh_message(self) -> MXSuperRefreshMessage:
        super_refresh_msg = MXSuperRefreshMessage(self)
        return super_refresh_msg

    @override
    def _get_function(self, function_name: str) -> MXSuperFunction:
        for function in self.function_list:
            if function.name == function_name:
                return function

    @override
    def _subscribe_init_topic_list(self):
        super()._subscribe_init_topic_list()

        topic_list = [MXProtocolType.Super.MS_RESULT_SERVICE_LIST.value % "#"]

        for topic in topic_list:
            self._subscribe(topic)

    @override
    def _subscribe_service_topic_list(self):
        for function in self.function_list:
            # Super Schedule, Super Execute에 필요한 토픽들을 미리 구독을 해놓는다.
            topic_list = [
                MXProtocolType.Super.MS_SCHEDULE.value % (function.name, self._name, self._middleware_name, '#'),
                MXProtocolType.Super.MS_RESULT_SCHEDULE.value % ('+', '+', '#'),
                MXProtocolType.Super.MS_EXECUTE.value % (function.name, self._name, self._middleware_name, '#'),
                MXProtocolType.Super.MS_RESULT_EXECUTE.value % ('+', '+', '#'),
            ]

            for topic in topic_list:
                self._subscribe(topic)

    def _request_sub_service_execute(self, sub_service_execute_request: MXSubExecuteRequest) -> None:
        if not isinstance(sub_service_execute_request, MXSubExecuteRequest):
            raise MXTypeError(f'[{get_current_function_name()}] Invalid type of sub_service_execute_request: {type(sub_service_execute_request)}')

        sub_service_execute_msg = sub_service_execute_request.trigger_msg
        self._send_SM_EXECUTE(sub_service_execute_msg)

    # 하나의 sub_service_request에 있는 sub_service_execute_request들을 병렬로 실행한다.
    def _sub_service_execute_parallel(
        self, sub_service_execute_request_list: List[MXSubExecuteRequest], arg_list: List[Union[int, float, str, bool]]
    ) -> List[Union[int, float, str, bool]]:
        result_list = []
        for i, sub_service_execute_request in enumerate(sub_service_execute_request_list):
            sub_service_execute_request.timer_start()
            sub_service_execute_request.trigger_msg.arguments = tuple(arg_list)

            sub_service_execute_msg = sub_service_execute_request.trigger_msg
            MXLOG_DEBUG(
                f'[SUB_EXECUTE START] {sub_service_execute_msg.sub_service_name}|{sub_service_execute_msg.target_middleware_name}|{sub_service_execute_msg.scenario}|{i}',
                'cyan',
            )
            self._request_sub_service_execute(sub_service_execute_request)

        for i, sub_service_execute_request in enumerate(sub_service_execute_request_list):
            sub_service_execute_request.result_msg = sub_service_execute_request.get_result_msg()
            sub_service_execute_request.timer_end()
            rc_msg = sub_service_execute_request.result_msg

            sub_service_execute_msg: MXSubExecuteMessage = sub_service_execute_request.trigger_msg
            sub_service_execute_result_msg: MXSubExecuteResultMessage = sub_service_execute_request.result_msg
            result_list.append(
                dict(
                    scenario=sub_service_execute_result_msg.scenario,
                    return_type=sub_service_execute_result_msg.return_type,
                    return_value=sub_service_execute_result_msg.return_value,
                    error=sub_service_execute_result_msg.error,
                )
            )

            MXLOG_DEBUG(
                f'[SUB_EXECUTE END] {sub_service_execute_msg.sub_service_name}|{sub_service_execute_msg.target_middleware_name}|{sub_service_execute_msg.scenario}|{i}'
                f'duration: {sub_service_execute_request.duration():.4f} Sec',
                'cyan',
            )

        return result_list

    def _get_sub_service_from_global_service_table(self, sub_service_name: str) -> MXFunction:
        for function in self._global_service_table['functions']:
            if function.name == sub_service_name:
                return function
        return None

    def _check_sub_service_callable(self, sub_service_name: str, return_type: MXType):
        global_sub_service = self._get_sub_service_from_global_service_table(sub_service_name)
        if not global_sub_service:
            return False
        if not global_sub_service.return_type == return_type:
            return False

        return True

    def _check_req_valid(
        self,
        sub_service_name: str,
        tag_list: List[str],
        arg_list: Union[Tuple[MXArgument], Tuple],
        return_type: MXType,
        service_type: MXServiceType,
        range_type: MXRangeType,
    ):
        if not sub_service_name:
            raise MXValueError(f'sub_service_name must be not empty')
        if not tag_list:
            raise MXValueError(f'tag_list must be not empty')
        if not all(tag_list):
            raise MXValueError(f'tag in tag_list must be not empty string')
        if not service_type in [MXServiceType.VALUE, MXServiceType.FUNCTION]:
            raise MXTypeError(f'Invalid service_type: {service_type}')
        if not return_type in [MXType.INTEGER, MXType.DOUBLE, MXType.STRING, MXType.BOOL, MXType.BINARY, MXType.VOID]:
            raise MXTypeError(f'Invalid return_type: {return_type}')
        elif service_type == MXServiceType.VALUE and return_type == MXType.VOID:
            raise MXTypeError(f'Value service cannot have a return_type of void')
        if not range_type in [MXRangeType.SINGLE, MXRangeType.ALL]:
            raise MXTypeError(f'Invalid range_type: {range_type}')

        return True

    def _check_req_return_type(self, sub_service_return_type: MXType, req_return_type: MXType, service_type: MXServiceType):
        if req_return_type in [MXType.INTEGER, MXType.DOUBLE] and sub_service_return_type in [
            MXType.INTEGER,
            MXType.DOUBLE,
        ]:
            return True
        elif req_return_type == MXType.VOID and service_type == MXServiceType.VALUE:
            raise MXTypeError(f'Not matched return_type. Value service cannot have a return_type of void: {sub_service_return_type}')
        elif req_return_type != sub_service_return_type:
            raise MXTypeError(f'Not matched return_type: {sub_service_return_type} != {req_return_type}')

        return True

    def _check_super_service_exist(self, super_service: MXSuperFunction):
        return all(
            [
                self._check_sub_service_callable(
                    sub_service_request._sub_service_type.name,
                    sub_service_request._sub_service_type.return_type,
                )
                for sub_service_request in super_service._sub_service_request_list
            ]
        )

    def req(
        self,
        sub_service_name: str,
        tag_list: List[str],
        arg_list: Union[Tuple[MXArgument], Tuple] = [],
        return_type: MXType = MXType.UNDEFINED,
        service_type: MXServiceType = MXServiceType.FUNCTION,
        range_type: MXRangeType = MXRangeType.SINGLE,
    ) -> Union[List[dict], bool]:
        # Detect fatal errors.
        # If an error occurs, the program terminates by raising an exception.
        if not self._check_req_valid(
            sub_service_name=sub_service_name,
            tag_list=tag_list,
            arg_list=arg_list,
            return_type=return_type,
            service_type=service_type,
            range_type=range_type,
        ):
            return False

        super_service_name = get_upper_function_name()
        target_super_service = self._get_function(super_service_name)
        target_sub_service = self._get_sub_service_from_global_service_table(sub_service_name)

        # Convert tag of type [str] to [MXTag]
        tag_list = [MXTag(str_tag) for str_tag in tag_list]

        if service_type == MXServiceType.VALUE:
            sub_service_name = f'__{sub_service_name}'
        elif service_type == MXServiceType.FUNCTION:
            sub_service_name = sub_service_name

        # When initiate a super thing, extract information about the super service.
        if not target_super_service.get_is_scanned():
            target_super_service.add_sub_service_request_info(
                sub_service_name=sub_service_name,
                arg_list=arg_list,
                tag_list=tag_list,
                return_type=return_type,
                range_type=range_type,
            )
            return []
        else:
            if not self._check_sub_service_callable(sub_service_name, return_type):
                MXLOG_DEBUG(f'sub_service {sub_service_name} is not callable', 'red')
                return False
            if not self._compare_arg_list(target_sub_service.arg_list, list(arg_list)):
                MXLOG_DEBUG(f'Not matched arg_list')
                return False
            if not self._check_req_return_type(target_sub_service.return_type, req_return_type=return_type, service_type=service_type):
                MXLOG_DEBUG(f'Not matched return_type')
                return False

            current_thread = threading.current_thread()
            scenario_name = current_thread.user_data['scenario']
            requester_middleware_name = current_thread.user_data['requester_middleware']
            # super_service_request_key = scenario_name@requester_middleware_name
            super_service_request_key = make_super_request_key(scenario_name, requester_middleware_name)

            MXLOG_DEBUG(
                f'[DEBUG] before pop SUPER_SERVICE_REQUEST_KEY_TABLE: \n{dict_to_json_string(self._SUPER_SERVICE_REQUEST_KEY_TABLE, pretty=True)}'
                f'\n super_service_request_key: {super_service_request_key} '
            )
            sub_service_request_key_list = self._SUPER_SERVICE_REQUEST_KEY_TABLE[super_service_request_key]
            # sub_service_request_key = sub_service_name@sub_service_request_order
            sub_service_request_key = sub_service_request_key_list.pop(0)
            MXLOG_DEBUG(
                f'[DEBUG] after pop SUPER_SERVICE_REQUEST_KEY_TABLE: \n{dict_to_json_string(self._SUPER_SERVICE_REQUEST_KEY_TABLE, pretty=True)}'
                f'\n super_service_request_key: {super_service_request_key} '
            )
            if len(sub_service_request_key_list) == 0:
                self._SUPER_SERVICE_REQUEST_KEY_TABLE.pop(super_service_request_key)

            super_service_execute_request = target_super_service._mapping_table[super_service_request_key]
            sub_service_execute_request_list = super_service_execute_request._sub_service_request_table[sub_service_request_key]._target_request_list

            result_list = self._sub_service_execute_parallel(sub_service_execute_request_list, arg_list)
            return result_list

    # TODO: implement this
    def r(self, line: str = None, *arg_list) -> Union[List[dict], bool]:
        super_service_name = get_upper_function_name()

        range_type = 'all' if 'all' in line else 'single'
        function_name = line.split('.')[1][0: line.split('.')[1].find('(')]
        bracket_parse: List[str] = re.findall(r'\(.*?\)', line)
        tags = [tag[1:] for tag in bracket_parse[0][1:-1].split(' ')]

        arguments = []
        for bracket_inner_element in bracket_parse[1][1:-1].split(','):
            bracket_inner_element = bracket_inner_element.strip(' ')
            if bracket_inner_element == '':
                continue
            else:
                arguments.append(bracket_inner_element)

        for i, arg in enumerate(arguments):
            if '$' in arg:
                index = int(arg[1:])
                arguments[i] = arg_list[index - 1]

        arguments = tuple(arguments)

    def _extract_sub_service_request_info(self) -> None:
        for function in self.function_list:
            if self.is_super and not function.get_is_scanned():
                arg_list = function.arg_list
                try:
                    MXLOG_DEBUG(f'Detect super service: {function.name}', 'green')
                    function._func(*tuple(arg_list))
                except MXError as e:
                    # req를 실행하다가 MySSIXError와 관련된 에러가 발생한다는 것은 req명세가 잘못
                    # 되었다는 것을 의미한다. 만약 MySSIXError에러가 아닌 다른 예외가 발생한 경우,
                    # super service안에 있는 코드 중, req() 코드 부분이 아닌 코드에 의한 예외이므로
                    # 정상적으로 req()에 대한 정보를 추출한 것이다.
                    raise e
                else:
                    function.set_is_scanned(True)

    def _extract_value_info(self, thing: dict, middleware_name: str) -> List[MXValue]:
        thing_name = thing['name']
        value_list = thing['values']

        value_service_list = []
        for value_info in value_list:
            value_tag_list = [MXTag(tag['name']) for tag in value_info['tags']]

            # TODO: cycle info is omit in service list
            value_service = MXValue(
                func=dummy_func(arg_list=[]),
                type=MXType.get(value_info['type']),
                bound=(float(value_info['bound']['min_value']), float(value_info['bound']['max_value'])),
                cycle=None,
                name=value_info['name'],
                tag_list=value_tag_list,
                desc=value_info['description'],
                thing_name=thing_name,
                middleware_name=middleware_name,
                format=value_info['format'] if not '(null)' in value_info['format'] else '',
            )
            if value_service not in self._global_service_table['values']:
                value_service_list.append(value_service)

        return value_service_list

    def _extract_function_info(self, thing_info: dict, middleware_name: str) -> List[MXFunction]:
        thing_name = thing_info['name']
        function_list = thing_info['functions']

        function_service_list = []
        for function_info in function_list:
            function_tag_list = [MXTag(tag['name']) for tag in function_info['tags']]
            arg_list = []
            if bool(function_info['use_arg']):
                for argument in function_info['arguments']:
                    arg_list.append(
                        MXArgument(
                            name=argument['name'],
                            type=MXType.get(argument['type']),
                            bound=(float(argument['bound']['min_value']), float(argument['bound']['max_value'])),
                        )
                    )

            function_service = MXFunction(
                func=dummy_func(arg_list=arg_list),
                return_type=MXType.get(function_info['return_type']),
                name=function_info['name'],
                tag_list=function_tag_list,
                desc=function_info['description'],
                thing_name=thing_name,
                middleware_name=middleware_name,
                arg_list=arg_list,
                exec_time=function_info['exec_time'],
            )
            if function_service not in self._global_service_table['functions']:
                function_service_list.append(function_service)

        return function_service_list

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @override
    @property
    def function_list(self) -> List[MXSuperFunction]:
        return sorted([service for service in self._service_list if isinstance(service, MXFunction)], key=lambda x: x.name)

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    @override
    def set_function_list(self, function_list: List[MXSuperFunction]) -> None:
        self.function_list = function_list



from big_thing_py.core.service_model import Objects as Skill
from big_thing_py.core.device_model.DeviceObjects import MXDeviceCategory
from typing import Set, Dict, List, Union, Any


class AirConditioner(MXDeviceCategory, device_type=0x0000):
    skills = {
        Skill.switch,
        Skill.airConditionerMode,
    }


class AirPurifier(MXDeviceCategory, device_type=0x0001):
    skills = {
        Skill.switch,
        Skill.airPurifierFanMode,
    }


class AirQualityDetector(MXDeviceCategory, device_type=0x0002):
    skills = {
        Skill.dustSensor,
        Skill.carbonDioxideMeasurement,
        Skill.temperatureMeasurement,
        Skill.relativeHumidityMeasurement,
        Skill.tvocMeasurement,
    }


class Blind(MXDeviceCategory, device_type=0x0003):
    skills = {
        Skill.blind,
        Skill.blindLevel,
    }


class Button(MXDeviceCategory, device_type=0x0004):
    skills = {
        Skill.button,
    }


class Calculator(MXDeviceCategory, device_type=0x0005):
    skills = {
        Skill.calculator,
    }


class Camera(MXDeviceCategory, device_type=0x0006):
    skills = {
        Skill.switch,
        Skill.camera,
    }


class Charger(MXDeviceCategory, device_type=0x0007):
    skills = {
        Skill.chargingState,
        Skill.currentMeasurement,
        Skill.voltageMeasurement,
    }


class Clock(MXDeviceCategory, device_type=0x0008):
    skills = {
        Skill.clock,
    }


class ContactSensor(MXDeviceCategory, device_type=0x0009):
    skills = {
        Skill.contactSensor,
    }


class Curtain(MXDeviceCategory, device_type=0x000A):
    skills = {
        Skill.curtain,
    }


class Dehumidifier(MXDeviceCategory, device_type=0x0010):
    skills = {
        Skill.switch,
        Skill.dehumidifierMode,
    }


class Dishwasher(MXDeviceCategory, device_type=0x0011):
    skills = {
        Skill.switch,
        Skill.dishwasherMode,
    }


class DoorLock(MXDeviceCategory, device_type=0x0012):
    skills = {
        Skill.doorControl,
    }


class EmailProvider(MXDeviceCategory, device_type=0x0013):
    skills = {
        Skill.emailProvider,
    }


class Fan(MXDeviceCategory, device_type=0x0014):
    skills = {
        Skill.switch,
        Skill.fanControl,
    }


class Feeder(MXDeviceCategory, device_type=0x0015):
    skills = {
        Skill.switch,
        Skill.feederOperatingState,
        Skill.feederPortion,
    }


class GasMeter(MXDeviceCategory, device_type=0x0016):
    skills = {
        Skill.gasMeter,
    }


class GasValve(MXDeviceCategory, device_type=0x0017):
    skills = {
        Skill.gasMeter,
        Skill.valve,
    }


class Humidifier(MXDeviceCategory, device_type=0x0018):
    skills = {
        Skill.switch,
        Skill.humidifierMode,
    }


class HumiditySensor(MXDeviceCategory, device_type=0x0019):
    skills = {
        Skill.relativeHumidityMeasurement,
    }


class Irrigator(MXDeviceCategory, device_type=0x001A):
    skills = {
        Skill.switch,
        Skill.irrigatorOperatingState,
        Skill.irrigatorPortion,
    }


class LeakSensor(MXDeviceCategory, device_type=0x001B):
    skills = {
        Skill.leakSensor,
    }


class Light(MXDeviceCategory, device_type=0x001C):
    skills = {
        Skill.switch,
        Skill.switchLevel,
        Skill.colorControl,
    }


class LightSensor(MXDeviceCategory, device_type=0x001D):
    skills = {
        Skill.lightLevel,
    }


class MenuProvider(MXDeviceCategory, device_type=0x001E):
    skills = {
        Skill.menuProvider,
    }


class MotionSensor(MXDeviceCategory, device_type=0x001F):
    skills = {
        Skill.motionSensor,
    }


class PresenceSensor(MXDeviceCategory, device_type=0x0020):
    skills = {
        Skill.presenceSensor,
    }


class Pump(MXDeviceCategory, device_type=0x0021):
    skills = {
        Skill.switch,
        Skill.pump,
        Skill.pumpOperationMode,
    }


class Refrigerator(MXDeviceCategory, device_type=0x0022):
    skills = {
        Skill.switch,
        Skill.refrigeration,
    }


class RobotCleaner(MXDeviceCategory, device_type=0x0023):
    skills = {
        Skill.switch,
        Skill.robotCleanerCleaningMode,
    }


class Shade(MXDeviceCategory, device_type=0x0024):
    skills = {
        Skill.windowShade,
        Skill.windowShadeLevel,
    }


class Siren(MXDeviceCategory, device_type=0x0025):
    skills = {
        Skill.switch,
        Skill.sirenMode,
    }


class SmartPlug(MXDeviceCategory, device_type=0x0026):
    skills = {
        Skill.switch,
        Skill.powerMeter,
        Skill.voltageMeasurement,
        Skill.currentMeasurement,
    }


class SmokeDetector(MXDeviceCategory, device_type=0x0027):
    skills = {
        Skill.smokeDetector,
    }


class SoundSensor(MXDeviceCategory, device_type=0x0028):
    skills = {
        Skill.soundSensor,
        Skill.soundPressureLevel,
    }


class Speaker(MXDeviceCategory, device_type=0x0029):
    skills = {
        Skill.switch,
        Skill.mediaPlayback,
    }


class Switch(MXDeviceCategory, device_type=0x002A):
    skills = {
        Skill.switch,
    }


class Television(MXDeviceCategory, device_type=0x002C):
    skills = {
        Skill.switch,
        Skill.tvChannel,
        Skill.audioMute,
        Skill.audioVolume,
    }


class TemperatureSensor(MXDeviceCategory, device_type=0x002D):
    skills = {
        Skill.temperatureMeasurement,
    }


class TestDevice(MXDeviceCategory, device_type=0x002B):
    skills = {
        Skill.testSkill,
    }


class Valve(MXDeviceCategory, device_type=0x002E):
    skills = {
        Skill.valve,
    }


class WeatherProvider(MXDeviceCategory, device_type=0x002F):
    skills = {
        Skill.weatherProvider,
    }


class Window(MXDeviceCategory, device_type=0x0030):
    skills = {
        Skill.windowControl,
    }


class FaceRecognizer(MXDeviceCategory, device_type=0x0031):
    skills = {}


class CloudServiceProvider(MXDeviceCategory, device_type=0x0032):
    skills = {}
    
class FallDetector(MXDeviceCategory, device_type=0x0033):
    skills = {}
    
class NewsProvider(MXDeviceCategory, device_type=0x0034):
    skills = {}


####


class ManagerThing(MXDeviceCategory, device_type=0xF000):
    skills = {
        Skill.manager,
    }


class Undefined(MXDeviceCategory, device_type=0xFFFF):
    skills = {}


if __name__ == '__main__':
    print(Window.skills)



import typing
from enum import Enum
from dataclasses import dataclass, field
from big_thing_py.common import MXType
import sys
from typing import List, Union

INFINITY = sys.maxsize


class MXClassproperty:
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, instance, owner):
        return self.fget(owner)


ALL_SKILLS = {}


@dataclass
class MXValueType:
    type: MXType
    bound: List[int] = field(default_factory=lambda: [-INFINITY - 1, INFINITY])
    format: Union[str, Enum] = None


@dataclass
class Skill:
    id: str = field(init=False)

    @MXClassproperty
    def descriptor(cls) -> str:
        raise NotImplementedError()

    @MXClassproperty
    def values(cls) -> List['SkillValue']:
        value_list = []
        if not hasattr(cls, 'Values'):
            return value_list

        for attr in dir(cls.Values):
            if not attr.startswith('__'):
                attr_instance = getattr(cls.Values, attr)
                value_list.append(attr_instance)

        return value_list

    @MXClassproperty
    def functions(cls) -> List['SkillFunction']:
        function_list = []
        if not hasattr(cls, 'Functions'):
            return function_list

        for attr in dir(cls.Functions):
            if not attr.startswith('__'):
                attr_instance = getattr(cls.Functions, attr)
                function_list.append(attr_instance)

        return function_list

    @MXClassproperty
    def docs_string(cls) -> str:
        description = f'{cls.id}\n'
        description += f'\ndescription: {cls.descriptor}\n'
        if hasattr(cls, 'Values'):
            description += '\nValues:\n'
            for value in cls.values:
                description += f'Value [{value.value_id}]\n'
                description += f'  - {value.docs_string}\n'
        if hasattr(cls, 'Functions'):
            description += '\nFunctions:\n'
            for function in cls.functions:
                description += f'  - {function.docs_string}\n'
        return description

    def __init_subclass__(cls, *args, **kwargs) -> None:
        '''Register a subclass.'''
        super().__init_subclass__(*args, **kwargs)
        cls.id = cls.__name__
        # register this cluster in the ALL_CLUSTERS dict for quick lookups
        try:
            ALL_SKILLS[cls.id] = cls
        except NotImplementedError:
            # handle case where the Cluster class is not (fully) subclassed
            # and accessing the id property throws a NotImplementedError.
            pass


class SkillValue:
    @MXClassproperty
    def descriptor(cls) -> str:
        raise NotImplementedError()

    @MXClassproperty
    def skill_id(cls) -> str:
        return cls.__qualname__.split('.')[0]

    @MXClassproperty
    def value_id(cls) -> str:
        return cls.__name__

    @MXClassproperty
    def value_type(cls) -> MXValueType:
        raise NotImplementedError()

    @MXClassproperty
    def docs_string(cls) -> str:
        return f'Value: {cls.descriptor} (Type: {str(cls.value_type.type)}, Bound: {cls.value_type.bound}, Format: {cls.value_type.format.__name__})'


class SkillFunctionArgument:
    @MXClassproperty
    def descriptor(cls) -> str:
        raise NotImplementedError()

    @MXClassproperty
    def function_id(cls) -> str:
        return cls.__qualname__.split('.')[2]

    @MXClassproperty
    def argument_id(cls) -> str:
        return cls.__name__

    @MXClassproperty
    def argument_type(cls) -> MXType:
        raise NotImplementedError()

    @MXClassproperty
    def docs_string(cls) -> str:
        return f'Argument: {cls.descriptor} (Type: {str(cls.argument_type.type)}, Format: {cls.format_details()})'


class SkillFunction:
    @MXClassproperty
    def descriptor(cls) -> str:
        raise NotImplementedError()

    @MXClassproperty
    def skill_id(cls) -> str:
        return cls.__qualname__.split('.')[0]

    @MXClassproperty
    def function_id(cls) -> str:
        return cls.__name__

    @MXClassproperty
    def arguments(cls) -> List[SkillFunctionArgument]:
        argument_list = []
        if not hasattr(cls, 'Arguments'):
            return argument_list

        for attr in dir(cls.Arguments):
            if not attr.startswith('__'):
                attr_instance = getattr(cls.Arguments, attr)
                argument_list.append(attr_instance)

        return argument_list

    @MXClassproperty
    def return_type(cls) -> MXType:
        raise NotImplementedError()

    @MXClassproperty
    def docs_string(cls) -> str:
        args = ', '.join([arg.docs_string for arg in cls.arguments()])
        return f'Function: {cls.descriptor} - Arguments: ({args}) - Returns: {cls.return_type().name}'



from enum import Enum, auto
from typing import Union
import base64
import re


class MXNetworkType(Enum):
    MQTT = 0
    API = 1
    BLUETOOTH = 2
    BLE = 3
    RF = 4
    XBEE = 5


class BinaryBase64:
    def __init__(self, input: str) -> None:
        self._data = base64.b64encode(input.encode()).decode('utf-8')

    def __str__(self) -> str:
        return self._data

    def __repr__(self) -> str:
        return self._data

    @property
    def binary(self) -> str:
        return self._data

    @property
    def text(self) -> str:
        return base64.b64decode(self._data).decode('utf-8')

    @classmethod
    def from_normal_string(cls, s: str) -> 'BinaryBase64':
        obj = cls.__new__(cls)
        obj._data = base64.b64encode(s.encode()).decode('utf-8')
        return obj

    @classmethod
    def from_binary_string(cls, b: str) -> 'BinaryBase64':
        if cls.is_base64(b):
            obj = cls.__new__(cls)
            obj._data = b
            return obj
        else:
            raise ValueError('Input string is not a valid Base64 string.')

    def to_file(self, file_path: str) -> None:
        with open(file_path, 'wb') as f:
            f.write(base64.b64decode(self._data))

    @classmethod
    def from_file(cls, file_path: str) -> Union['BinaryBase64', None]:
        with open(file_path, 'rb') as f:
            return cls(base64.b64encode(f.read()).decode())

    @staticmethod
    def is_base64(base64_str: str) -> bool:
        '''
        NOTE: The minimum length set to 24 characters is chosen based on common use cases,
            such as encoded hashes or small files, which typically result in a Base64
            string longer than 24 characters. This length helps reduce the chance of
            mistaking regular short strings for Base64 encoded data.
        '''
        if len(base64_str) < 24 or len(base64_str) % 4 != 0:
            return False

        pattern = re.compile(r'^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$')
        try:
            base64.b64decode(base64_str, validate=True)
            return bool(pattern.match(base64_str))
        except Exception:
            return False


class MXType(Enum):
    def _generate_next_value_(name: str, start, count, last_values):
        return name.lower()

    UNDEFINED = auto()
    INTEGER = 'int'
    DOUBLE = 'double'
    STRING = 'string'
    BOOL = 'bool'
    BINARY = 'binary'
    ENUM = 'enum'
    LIST = 'list'
    DICT = 'dict'
    VOID = 'void'

    def __str__(self):
        return str(self.name)

    @classmethod
    def get(cls, name: Union[str, type]) -> 'MXType':
        try:
            if isinstance(name, str):
                name = name.lower()
                if name is not None:
                    for mxtype in MXType:
                        if mxtype.value == name:
                            return mxtype
                        elif mxtype.value == 'int' and name == 'integer':
                            return cls.INTEGER
                        elif mxtype.value == 'double' and name == 'float':
                            return cls.DOUBLE
                        elif mxtype.value == 'string' and name == 'str':
                            return cls.STRING
                        elif mxtype.value == 'binary' and name == 'base64':
                            return cls.BINARY
                        elif mxtype.value == 'bool' and name == 'boolean':
                            return cls.BOOL
                    return cls.UNDEFINED
                else:
                    return cls.UNDEFINED
            elif isinstance(name, type):
                if name == int:
                    return MXType.INTEGER
                elif name == float:
                    return MXType.DOUBLE
                elif name == bool:
                    return MXType.BOOL
                elif name == str:
                    return MXType.STRING
                elif name == BinaryBase64:
                    return MXType.BINARY
                elif name is None:
                    return MXType.VOID
            elif isinstance(name, BinaryBase64):
                return MXType.BINARY
            elif name is None:
                return cls.VOID
        except Exception:
            return cls.UNDEFINED

    def to_pytype(self):
        if self == MXType.INTEGER:
            return int
        elif self == MXType.DOUBLE:
            return float
        elif self == MXType.BOOL:
            return bool
        elif self == MXType.STRING:
            return str
        elif self == MXType.BINARY:
            return BinaryBase64
        elif self == MXType.VOID:
            return None


class MXActionType(Enum):
    def _generate_next_value_(name: str, start, count, last_values):
        return name.lower()

    REGISTER = auto()
    EXECUTE = auto()
    ALIVE = auto()
    VALUE_PUBLISH = auto()
    REFRESH = auto()
    SUPER_SCHEDULE = auto()
    SUPER_EXECUTE = auto()
    SUB_SCHEDULE = auto()
    SUB_EXECUTE = auto()
    INNER_EXECUTE = auto()


class MXProtocolType:
    class Base(Enum):
        # ===============
        # ___  ___ _____
        # |  \/  ||_   _|
        # | .  . |  | |
        # | |\/| |  | |
        # | |  | |  | |
        # \_|  |_/  \_/
        # ===============

        # MT/REQUEST_REGISTER_INFO/[ThingName]
        MT_REQUEST_REGISTER_INFO = 'MT/REQUEST_REGISTER_INFO/%s'

        # MT/REQUEST_UNREGISTER/[ThingName]
        MT_REQUEST_UNREGISTER = 'MT/REQUEST_UNREGISTER/%s'

        # MT/RESULT/REGISTER/[ThingName]
        MT_RESULT_REGISTER = 'MT/RESULT/REGISTER/%s'

        # MT/RESULT/UNREGISTER/[ThingName]
        MT_RESULT_UNREGISTER = 'MT/RESULT/UNREGISTER/%s'

        # MT/EXECUTE/[FunctionName]/[ThingName]/([TargetMiddlewareName]/[Request_ID])
        # Request_ID = requester_middleware@super_thing@super_service@subrequest_order
        MT_EXECUTE = 'MT/EXECUTE/%s/%s/%s/%s'

        # MT/IN/EXECUTE/[FunctionName]/[ThingName]
        MT_IN_EXECUTE = 'MT/IN/EXECUTE/%s/%s'

        # MT/RESULT/BINARY_VALUE/[ThingName]
        MT_RESULT_BINARY_VALUE = 'MT/RESULT/BINARY_VALUE/%s'

        # ===============
        #  _____ ___  ___
        # |_   _||  \/  |
        #   | |  | .  . |
        #   | |  | |\/| |
        #   | |  | |  | |
        #   \_/  \_|  |_/
        # ===============

        # TM/RESULT/EXECUTE/[FunctionName]/[ThingName]/([TargetMiddlewareName]/[Request_ID])
        # Request_ID = requester_middleware@super_thing@super_service@subrequest_order
        TM_RESULT_EXECUTE = 'TM/RESULT/EXECUTE/%s/%s/%s/%s'

        # TM/IN/RESULT/EXECUTE/[FunctionName]/[ThingName]
        TM_IN_RESULT_EXECUTE = 'TM/IN/RESULT/EXECUTE/%s/%s'

        # TM/REGISTER/[ThingName]
        TM_REGISTER = 'TM/REGISTER/%s'

        # TM/UNREGISTER/[ThingName]
        TM_UNREGISTER = 'TM/UNREGISTER/%s'

        # TM/ALIVE/[ThingName]
        TM_ALIVE = 'TM/ALIVE/%s'

        # [ValueName]/[ThingName]
        TM_VALUE_PUBLISH = '%s/%s'

        def get_prefix(self):
            topic_tree = self.value.split('/')
            result_topic = []
            for topic_part in topic_tree:
                if topic_part != '%s':
                    result_topic.append(topic_part)

            return '/'.join(result_topic)

    class Super(Enum):
        # ===============
        # ___  ___ _____
        # |  \/  |/  ___|
        # | .  . |\ `--.
        # | |\/| | `--. \
        # | |  | |/\__/ /
        # \_|  |_/\____/
        # ===============

        # MS/SCHEDULE/[SuperFunctionName]/[SuperThingName]/[SuperMiddlewareName]/[RequesterMWName]
        MS_SCHEDULE = 'MS/SCHEDULE/%s/%s/%s/%s'

        # MS/EXECUTE/[SuperFunctionName]/[SuperThingName]/[SuperMiddlewareName]/[RequesterMWName]
        MS_EXECUTE = 'MS/EXECUTE/%s/%s/%s/%s'

        # MS/RESULT/SCHEDULE/[TargetFunctionName]/SUPER/[TargetMiddlewareName]/[Request_ID]
        # Request_ID = requester_middleware@super_thing@super_service@subrequest_order
        MS_RESULT_SCHEDULE = 'MS/RESULT/SCHEDULE/%s/SUPER/%s/%s'

        # MS/RESULT/EXECUTE/[TargetFunctionName]/SUPER/[TargetMiddlewareName]/[Request_ID]
        # Request_ID = requester_middleware@super_thing@super_service@subrequest_order
        MS_RESULT_EXECUTE = 'MS/RESULT/EXECUTE/%s/SUPER/%s/%s'

        # MS/RESULT/SERVICE_LIST/[SuperThingName]
        MS_RESULT_SERVICE_LIST = 'MS/RESULT/SERVICE_LIST/%s'

        # ================
        #  _____ ___  ___
        # /  ___||  \/  |
        # \ `--. | .  . |
        #  `--. \| |\/| |
        # /\__/ /| |  | |
        # \____/ \_|  |_/
        # ================

        # SM/SCHEDULE/[TargetFunctionName]/SUPER/[TargetMiddlewareName]/[Request_ID]
        # Request_ID = requester_middleware@super_thing@super_service@subrequest_order
        SM_SCHEDULE = 'SM/SCHEDULE/%s/SUPER/%s/%s'

        # SM/EXECUTE/[TargetFunctionName]/SUPER/[TargetMiddlewareName]/[Request_ID]
        # Request_ID = requester_middleware@super_thing@super_service@subrequest_order
        SM_EXECUTE = 'SM/EXECUTE/%s/SUPER/%s/%s'

        # SM/RESULT/SCHEDULE/[SuperFunctionName]/[SuperThingName]/[SuperMiddlewareName]/[RequesterMWName]
        SM_RESULT_SCHEDULE = 'SM/RESULT/SCHEDULE/%s/%s/%s/%s'

        # SM/RESULT/EXECUTE/[SuperFunctionName]/[SuperThingName]/[SuperMiddlewareName]/[RequesterMWName]
        SM_RESULT_EXECUTE = 'SM/RESULT/EXECUTE/%s/%s/%s/%s'

        # SM/REFRESH/[SuperThingName]
        SM_REFRESH = 'SM/REFRESH/%s'

        # ==================
        #   _____    _____
        #  |  __ \  / ____|
        #  | |__) || |
        #  |  ___/ | |
        #  | |     | |____
        #  |_|      \_____|
        # ==================

        # PC/SCHEDULE/[TargetFunctionName]/SUPER/[TargetMiddlewareName]/[Request_ID]
        PC_SCHEDULE = 'PC/SCHEDULE/%s/SUPER/%s/%s'

        # PC/EXECUTE/[TargetFunctionName]/SUPER/[TargetMiddlewareName]/[Request_ID]
        PC_EXECUTE = 'PC/EXECUTE/%s/SUPER/%s/%s'

        # PC/RESULT/SCHEDULE/[SuperFunctionName]/[SuperThingName]/[SuperMiddlewareName]/[RequesterMWName]
        PC_RESULT_SCHEDULE = 'PC/RESULT/SCHEDULE/%s/%s/%s/%s'

        # PC/RESULT/EXECUTE/[SuperFunctionName]/[SuperThingName]/[SuperMiddlewareName]/[RequesterMWName]
        PC_RESULT_EXECUTE = 'PC/RESULT/EXECUTE/%s/%s/%s/%s'

        # PC/SERVICE_LIST/#
        PC_SERVICE_LIST = 'PC/SERVICE_LIST/%s'

        # PC/TRAVERSE/#
        PC_TRAVERSE = 'PC/TRAVERSE/%s'

        # ==================
        #    _____  _____
        #   / ____||  __ \
        #  | |     | |__) |
        #  | |     |  ___/
        #  | |____ | |
        #   \_____||_|
        # ==================

        # CP/SCHEDULE/[SuperFunctionName]/[SuperThingName]/[SuperMiddlewareName]/[RequesterMWName]
        CP_SCHEDULE = 'CP/SCHEDULE/%s/%s/%s/%s'

        # CP/EXECUTE/[SuperFunctionName]/[SuperThingName]/[SuperMiddlewareName]/[Requester MWName]
        CP_EXECUTE = 'CP/EXECUTE/%s/%s/%s/%s'

        # CP/RESULT/SCHEDULE/[TargetFunctionName]/SUPER/[TargetMiddlewareName]/[Request_ID]
        CP_RESULT_SCHEDULE = 'CP/RESULT/SCHEDULE/%s/SUPER/%s/%s'

        # CP/RESULT/EXECUTE/[TargetFunctionName]/SUPER/[TargetMiddlewareName]/[Request_ID]
        CP_RESULT_EXECUTE = 'CP/RESULT/EXECUTE/%s/SUPER/%s/%s'

        # CP/SERVICE_LIST/#
        CP_SERVICE_LIST = 'CP/SERVICE_LIST/%s'

        def get_prefix(self):
            topic_tree = self.value.split('/')
            result_topic = []
            for topic_part in topic_tree:
                if topic_part != '%s':
                    result_topic.append(topic_part)

            return '/'.join(result_topic)

    class WebClient(Enum):
        # ==================
        #   ______  __  __
        #  |  ____||  \/  |
        #  | |__   | \  / |
        #  |  __|  | |\/| |
        #  | |____ | |  | |
        #  |______||_|  |_|
        # ==================

        # Hub Connect
        # EM/REQUEST_CONNECTION/[ClientID]
        EM_REQUEST_CONNECTION = 'EM/REQUEST_CONNECTION/%s'

        # EM/CONNECT_AS_OWNER/[ClientID]
        EM_CONNECT_AS_OWNER = 'EM/CONNECT_AS_OWNER/%s'

        # Member Manage
        # EM/SET_MEMBER_LEVEL/[ClientID]
        EM_SET_MEMBER_LEVEL = 'EM/SET_MEMBER_LEVEL/%s'

        # EM/SET_MEMBER_STATE/[ClientID]
        EM_SET_MEMBER_STATE = 'EM/SET_MEMBER_STATE/%s'

        # EM/DELETE_MEMBER/[ClientID]
        EM_DELETE_MEMBER = 'EM/DELETE_MEMBER/%s'

        # EM/ADD_THING_FAVORITE/[ClientID]
        EM_ADD_THING_FAVORITE = 'EM/ADD_THING_FAVORITE/%s'

        # EM/DELETE_THING_FAVORITE/[ClientID]
        EM_DELETE_THING_FAVORITE = 'EM/DELETE_THING_FAVORITE/%s'

        # EM/ADD_SERVICE_FAVORITE/[ClientID]
        EM_ADD_SERVICE_FAVORITE = 'EM/ADD_SERVICE_FAVORITE/%s'

        # EM/DELETE_SERVICE_FAVORITE/[ClientID]
        EM_DELETE_SERVICE_FAVORITE = 'EM/DELETE_SERVICE_FAVORITE/%s'

        # EM/ADD_SCENARIO_FAVORITE/[ClientID]
        EM_ADD_SCENARIO_FAVORITE = 'EM/ADD_SCENARIO_FAVORITE/%s'

        # EM/DELETE_SCENARIO_FAVORITE/[ClientID]
        EM_DELETE_SCENARIO_FAVORITE = 'EM/DELETE_SCENARIO_FAVORITE/%s'

        # Device Manage
        # EM/REGISTER_THING/[ClientID]
        EM_REGISTER_THING = 'EM/REGISTER_THING/%s'

        # EM/UNREGISTER_THING/[ClientID]
        EM_UNREGISTER_THING = 'EM/UNREGISTER_THING/%s'

        # EM/SET_NICK_NAME/[ClientID]
        EM_SET_NICK_NAME = 'EM/SET_NICK_NAME/%s'

        # EM/SET_DESCRIPTION/[ClientID]
        EM_SET_DESCRIPTION = 'EM/SET_DESCRIPTION/%s'

        # EM/REQUEST_SCAN/[ClientID]
        EM_REQUEST_SCAN = 'EM/REQUEST_SCAN/%s'

        # EM/OPEN_WINDOW_MATTER/[ClientID]
        EM_OPEN_WINDOW_MATTER = 'EM/OPEN_WINDOW_MATTER/%s'

        # EM/CLOSE_WINDOW_MATTER/[ClientID]
        EM_CLOSE_WINDOW_MATTER = 'EM/CLOSE_WINDOW_MATTER/%s'

        # Scenario Manage
        # EM/VERIFY_SCENARIO/[ClientID]
        EM_VERIFY_SCENARIO = 'EM/VERIFY_SCENARIO/%s'

        # EM/ADD_SCENARIO/[ClientID]
        EM_ADD_SCENARIO = 'EM/ADD_SCENARIO/%s'

        # EM/RUN_SCENARIO/[ClientID]
        EM_RUN_SCENARIO = 'EM/RUN_SCENARIO/%s'

        # EM/STOP_SCENARIO/[ClientID]
        EM_STOP_SCENARIO = 'EM/STOP_SCENARIO/%s'

        # EM/UPDATE_SCENARIO/[ClientID]
        EM_UPDATE_SCENARIO = 'EM/UPDATE_SCENARIO/%s'

        # EM/MODIFY_SCENARIO/[ClientID]
        EM_MODIFY_SCENARIO = 'EM/MODIFY_SCENARIO/%s'

        # EM/DELETE_SCENARIO/[ClientID]
        EM_DELETE_SCENARIO = 'EM/DELETE_SCENARIO/%s'

        # Service Manage
        # EM/ADD_TAG/[ClientID]
        EM_ADD_TAG = 'EM/ADD_TAG/%s'

        # EM/DELETE_TAG/[ClientID]
        EM_DELETE_TAG = 'EM/DELETE_TAG/%s'

        # EM/SET_ACCESS/[ClientID]
        EM_SET_ACCESS = 'EM/SET_ACCESS/%s'

        # EM/SET_OPENED/[ClientID]
        EM_SET_OPENED = 'EM/SET_OPENED/%s'

        # EM/GET_TREE/[ClientID]
        EM_GET_TREE = 'EM/GET_TREE/%s'

        # EM/INSTANT_EXECUTE/[ClientID]
        EM_INSTANT_EXECUTE = 'EM/INSTANT_EXECUTE/%s'

        # Sync Info
        # EM/REFRESH/[ClientID]
        EM_REFRESH = 'EM/REFRESH/%s'

        # EM/GET_THING/[ClientID]
        EM_GET_THING = 'EM/GET_THING/%s'

        # EM/GET_SERVICE/[ClientID]
        EM_GET_SERVICE = 'EM/GET_SERVICE/%s'

        # EM/GET_SCENARIO/[ClientID]
        EM_GET_SCENARIO = 'EM/GET_SCENARIO/%s'

        # EM/GET_MEMBER/[ClientID]
        EM_GET_MEMBER = 'EM/GET_MEMBER/%s'

        # EM/GET_THING_LIST/[ClientID]
        EM_GET_THING_LIST = 'EM/GET_THING_LIST/%s'

        # EM/GET_SERVICE_LIST/[ClientID]
        EM_GET_SERVICE_LIST = 'EM/GET_SERVICE_LIST/%s'

        # EM/GET_SCENARIO_LIST/[ClientID]
        EM_GET_SCENARIO_LIST = 'EM/GET_SCENARIO_LIST/%s'

        # EM/GET_MEMBER_LIST/[ClientID]
        EM_GET_MEMBER_LIST = 'EM/GET_MEMBER_LIST/%s'

        # EM/SET_LOCATION/[ClientID]
        EM_SET_LOCATION = 'EM/SET_LOCATION/%s'

        # EM/GET_LOCATION/[ClientID]
        EM_GET_LOCATION = 'EM/GET_LOCATION/%s'

        # EM/GET_HOME/[ClientID]
        EM_GET_HOME = 'EM/GET_HOME/%s'

        # EM/GET_FAVORITES/[ClientID]
        EM_GET_FAVORITES = 'EM/GET_FAVORITES/%s'

        # Hub Reset
        # EM/RESET_HUB/[ClientID]
        EM_RESET_HUB = 'EM/RESET_HUB/%s'

        # ==================
        #   __  __  ______
        #  |  \/  ||  ____|
        #  | \  / || |__
        #  | |\/| ||  __|
        #  | |  | || |____
        #  |_|  |_||______|
        # ==================

        # Hub Connect
        # ME/RESULT/REQUEST_CONNECTION/[ClientID]
        ME_RESULT_REQUEST_CONNECTION = 'ME/RESULT/REQUEST_CONNECTION/%s'

        # ME/OWNER_BUTTON_CLICKED
        ME_OWNER_BUTTON_CLICKED = 'ME/OWNER_BUTTON_CLICKED'

        # ME/RESULT/CONNECT_AS_OWNER/[ClientID]
        ME_RESULT_CONNECT_AS_OWNER = 'ME/RESULT/CONNECT_AS_OWNER/%s'

        # Member Manage
        # ME/RESULT/SET_MEMBER_LEVEL/[ClientID]
        ME_RESULT_SET_MEMBER_LEVEL = 'ME/RESULT/SET_MEMBER_LEVEL/%s'

        # ME/RESULT/SET_MEMBER_STATE/[ClientID]
        ME_RESULT_SET_MEMBER_STATE = 'ME/RESULT/SET_MEMBER_STATE/%s'

        # ME/RESULT/DELETE_MEMBER/[ClientID]
        ME_RESULT_DELETE_MEMBER = 'ME/RESULT/DELETE_MEMBER/%s'

        # ME/RESULT/ADD_THING_FAVORITE/[ClientID]
        ME_RESULT_ADD_THING_FAVORITE = 'ME/RESULT/ADD_THING_FAVORITE/%s'

        # ME/RESULT/DELETE_THING_FAVORITE/[ClientID]
        ME_RESULT_DELETE_THING_FAVORITE = 'ME/RESULT/DELETE_THING_FAVORITE/%s'

        # ME/RESULT/ADD_SERVICE_FAVORITE/[ClientID]
        ME_RESULT_ADD_SERVICE_FAVORITE = 'ME/RESULT/ADD_SERVICE_FAVORITE/%s'

        # ME/RESULT/DELETE_SERVICE_FAVORITE/[ClientID]
        ME_RESULT_DELETE_SERVICE_FAVORITE = 'ME/RESULT/DELETE_SERVICE_FAVORITE/%s'

        # ME/RESULT/ADD_SCENARIO_FAVORITE/[ClientID]
        ME_RESULT_ADD_SCENARIO_FAVORITE = 'ME/RESULT/ADD_SCENARIO_FAVORITE/%s'

        # ME/RESULT/DELETE_SCENARIO_FAVORITE/[ClientID]
        ME_RESULT_DELETE_SCENARIO_FAVORITE = 'ME/RESULT/DELETE_SCENARIO_FAVORITE/%s'

        # Device Manage
        # ME/RESULT/REGISTER_THING/[ClientID]
        ME_RESULT_REGISTER_THING = 'ME/RESULT/REGISTER_THING/%s'

        # ME/RESULT/UNREGISTER/[ClientID]
        ME_RESULT_UNREGISTER_THING = 'ME/RESULT/UNREGISTER_THING/%s'

        # ME/RESULT/SET_NICK_NAME/[ClientID]
        ME_RESULT_SET_NICK_NAME = 'ME/RESULT/SET_NICK_NAME/%s'

        # ME/RESULT/SET_DESCRIPTION/[ClientID]
        ME_RESULT_SET_DESCRIPTION = 'ME/RESULT/SET_DESCRIPTION/%s'

        # ME/SCAN_RESULT/[ClientID]
        ME_SCAN_RESULT = 'ME/SCAN_RESULT/%s'

        # ME/RESULT/OPEN_WINDOW_MATTER/[ClientID]
        ME_RESULT_OPEN_WINDOW_MATTER = 'ME/RESULT/OPEN_WINDOW_MATTER/%s'

        # ME/RESULT/CLOSE_WINDOW_MATTER/[ClientID]
        ME_RESULT_CLOSE_WINDOW_MATTER = 'ME/RESULT/CLOSE_WINDOW_MATTER/%s'

        # Scenario Manage
        # ME/RESULT/VERIFY_SCENARIO/[ClientID]
        ME_RESULT_VERIFY_SCENARIO = 'ME/RESULT/VERIFY_SCENARIO/%s'

        # ME/RESULT/RUN_SCENARIO/[ClientID]
        ME_RESULT_RUN_SCENARIO = 'ME/RESULT/RUN_SCENARIO/%s'

        # ME/RESULT/STOP_SCENARIO/[ClientID]
        ME_RESULT_STOP_SCENARIO = 'ME/RESULT/STOP_SCENARIO/%s'

        # ME/RESULT/SCHEDULE_SCENARIO/[ClientID]
        ME_RESULT_SCHEDULE_SCENARIO = 'ME/RESULT/SCHEDULE_SCENARIO/%s'

        # ME/RESULT/SCHEDULE_SCENARIO/[ClientID]
        ME_RESULT_ADD_SCENARIO = ME_RESULT_SCHEDULE_SCENARIO

        # ME/RESULT/SCHEDULE_SCENARIO/[ClientID]
        ME_RESULT_UPDATE_SCENARIO = ME_RESULT_SCHEDULE_SCENARIO

        # ME/RESULT/SCHEDULE_SCENARIO/[ClientID]
        ME_RESULT_MODIFY_SCENARIO = ME_RESULT_SCHEDULE_SCENARIO

        # ME/RESULT/DELETE_SCENARIO/[ClientID]
        ME_RESULT_DELETE_SCENARIO = 'ME/RESULT/DELETE_SCENARIO/%s'

        # Service Manage
        # ME/RESULT/ADD_TAG/[ClientID]
        ME_RESULT_ADD_TAG = 'ME/RESULT/ADD_TAG/%s'

        # ME/RESULT/DELETE_TAG/[ClientID]
        ME_RESULT_DELETE_TAG = 'ME/RESULT/DELETE_TAG/%s'

        # ME/RESULT/SET_ACCESS/[ClientID]
        ME_RESULT_SET_ACCESS = 'ME/RESULT/SET_ACCESS/%s'

        # ME/RESULT/SET_OPENED/[ClientID]
        ME_RESULT_SET_OPENED = 'ME/RESULT/SET_OPENED/%s'

        # ME/RESULT/GET_TREE/[ClientID]
        ME_RESULT_GET_TREE = 'ME/RESULT/GET_TREE/%s'

        # ME/RESULT/INSTANT_EXECUTE/[ExecID]
        ME_RESULT_INSTANT_EXECUTE = 'ME/RESULT/INSTANT_EXECUTE/%s'

        # Sync Info
        # ME/RESULT/THING/[ClientID]
        ME_RESULT_THING = 'ME/RESULT/THING/%s'

        # ME/RESULT/SERVICE/[ClientID]
        ME_RESULT_SERVICE = 'ME/RESULT/SERVICE/%s'

        # ME/RESULT/SCENARIO/[ClientID]
        ME_RESULT_SCENARIO = 'ME/RESULT/SCENARIO/%s'

        # ME/RESULT/MEMBER/[ClientID]
        ME_RESULT_MEMBER = 'ME/RESULT/MEMBER/%s'

        # ME/RESULT/THING_LIST/[ClientID]
        ME_RESULT_THING_LIST = 'ME/RESULT/THING_LIST/%s'

        # ME/RESULT/SERVICE_LIST/[ClientID]
        ME_RESULT_SERVICE_LIST = 'ME/RESULT/SERVICE_LIST/%s'

        # ME/RESULT/SCENARIO_LIST/[ClientID]
        ME_RESULT_SCENARIO_LIST = 'ME/RESULT/SCENARIO_LIST/%s'

        # ME/RESULT/MEMBER_LIST/[ClientID]
        ME_RESULT_MEMBER_LIST = 'ME/RESULT/MEMBER_LIST/%s'

        # ME/RESULT/SET_LOCATION/[ClientID]
        ME_RESULT_SET_LOCATION = 'ME/RESULT/SET_LOCATION/%s'

        # ME/RESULT/GET_LOCATION/[ClientID]
        ME_RESULT_GET_LOCATION = 'ME/RESULT/GET_LOCATION/%s'

        # ME/RESULT/HOME/[ClientID]
        ME_RESULT_HOME = 'ME/RESULT/HOME/%s'

        # ME/RESULT/GET_FAVORITES/[ClientID]
        ME_RESULT_GET_FAVORITES = 'ME/RESULT/GET_FAVORITES/%s'

        # Notify
        # ME/NOTIFY_MESSAGE/[ClientID]
        ME_NOTIFY_MESSAGE = 'ME/NOTIFY_MESSAGE/%s'

        # ME/NOTIFY_CHANGE/[ClientID]
        ME_NOTIFY_CHANGE = 'ME/NOTIFY_CHANGE/%s'

        # Hub Reset
        # ME/RESULT/RESET_HUB/[ClientID]
        ME_RESULT_RESET_HUB = 'ME/RESULT/RESET_HUB/%s'

        def get_prefix(self):
            topic_tree = self.value.split('/')
            result_topic = []
            for topic_part in topic_tree:
                if topic_part != '%s':
                    result_topic.append(topic_part)

            return '/'.join(result_topic)

    class SN(Enum):
        # TM/SN/REGISTER/VALUE/[Thing ID]
        TM_SN_REGISTER_VALUE = "TM/SN/REGISTER/VALUE/%s"

        # TM/SN/REGISTER/VALUEDESC/[Thing ID]
        TM_SN_REGISTER_VALUEDESC = "TM/SN/REGISTER/VALUEDESC/%s"

        # TM/SN/REGISTER/VALUETAG/[Thing ID]
        TM_SN_REGISTER_VALUETAG = "TM/SN/REGISTER/VALUETAG/%s"

        # TM/SN/REGISTER/FUNCTION/[Thing ID]
        TM_SN_REGISTER_FUNCTION = "TM/SN/REGISTER/FUNCTION/%s"

        # TM/SN/REGISTER/FUNCTIONDESC/[Thing ID]
        TM_SN_REGISTER_FUNCTIONDESC = "TM/SN/REGISTER/FUNCTIONDESC/%s"

        # TM/SN/REGISTER/FUNCTIONTAG/[Thing ID]
        TM_SN_REGISTER_FUNCTIONTAG = "TM/SN/REGISTER/FUNCTIONTAG/%s"

        # TM/SN/REGISTER/ARGUMENT/[Thing ID]
        TM_SN_REGISTER_ARGUMENT = "TM/SN/REGISTER/ARGUMENT/%s"

        # TM/SN/REGISTER/ALIVECYCLE/[Thing ID]
        TM_SN_REGISTER_ALIVECYCLE = "TM/SN/REGISTER/ALIVECYCLE/%s"

        # TM/SN/REGISTER/FINISH/[Thing ID]
        TM_SN_REGISTER_FINISH = "TM/SN/REGISTER/FINISH/%s"

        def get_prefix(self):
            topic_tree = self.value.split('/')
            result_topic = []
            for topic_part in topic_tree:
                if topic_part != '%s':
                    result_topic.append(topic_part)

            return '/'.join(result_topic)

    @classmethod
    def get(cls, topic: str) -> 'MXProtocolType':
        # MT
        if 'MT/REQUEST_REGISTER_INFO/' in topic:
            return cls.Base.MT_REQUEST_REGISTER_INFO
        elif 'MT/REQUEST_UNREGISTER/' in topic:
            return cls.Base.MT_REQUEST_UNREGISTER
        elif 'MT/RESULT/REGISTER/' in topic:
            return cls.Base.MT_RESULT_REGISTER
        elif 'MT/RESULT/UNREGISTER/' in topic:
            return cls.Base.MT_RESULT_UNREGISTER
        elif 'MT/EXECUTE/' in topic:
            return cls.Base.MT_EXECUTE
        elif 'MT/IN/EXECUTE/' in topic:
            return cls.Base.MT_IN_EXECUTE
        elif 'MT/RESULT/BINARY_VALUE/' in topic:
            return cls.Base.MT_RESULT_BINARY_VALUE
        # TM
        elif 'TM/RESULT/EXECUTE/' in topic:
            return cls.Base.TM_RESULT_EXECUTE
        elif 'TM/IN/RESULT/EXECUTE/' in topic:
            return cls.Base.TM_IN_RESULT_EXECUTE
        elif 'TM/REGISTER/' in topic:
            return cls.Base.TM_REGISTER
        elif 'TM/UNREGISTER/' in topic:
            return cls.Base.TM_UNREGISTER
        elif 'TM/ALIVE/' in topic:
            return cls.Base.TM_ALIVE
        elif len(topic.split('/')) == 2:
            return cls.Base.TM_VALUE_PUBLISH
        # MS
        elif 'MS/SCHEDULE/' in topic:
            return cls.Super.MS_SCHEDULE
        elif 'MS/EXECUTE/' in topic:
            return cls.Super.MS_EXECUTE
        elif 'MS/RESULT/SCHEDULE/' in topic:
            return cls.Super.MS_RESULT_SCHEDULE
        elif 'MS/RESULT/EXECUTE/' in topic:
            return cls.Super.MS_RESULT_EXECUTE
        elif 'MS/RESULT/SERVICE_LIST/' in topic:
            return cls.Super.MS_RESULT_SERVICE_LIST
        # SM
        elif 'SM/SCHEDULE/' in topic:
            return cls.Super.SM_SCHEDULE
        elif 'SM/EXECUTE/' in topic:
            return cls.Super.SM_EXECUTE
        elif 'SM/RESULT/SCHEDULE/' in topic:
            return cls.Super.SM_RESULT_SCHEDULE
        elif 'SM/RESULT/EXECUTE/' in topic:
            return cls.Super.SM_RESULT_EXECUTE
        elif 'SM/REFRESH/' in topic:
            return cls.Super.SM_REFRESH
        # PC
        elif 'PC/SCHEDULE/' in topic:
            return cls.Super.PC_SCHEDULE
        elif 'PC/EXECUTE/' in topic:
            return cls.Super.PC_EXECUTE
        elif 'PC/RESULT/SCHEDULE/' in topic:
            return cls.Super.PC_RESULT_SCHEDULE
        elif 'PC/RESULT/EXECUTE/' in topic:
            return cls.Super.PC_RESULT_EXECUTE
        elif 'PC/SERVICE_LIST/' in topic:
            return cls.Super.PC_SERVICE_LIST
        elif 'PC/TRAVERSE/' in topic:
            return cls.Super.PC_TRAVERSE
        # CP
        elif 'CP/SCHEDULE/' in topic:
            return cls.Super.CP_SCHEDULE
        elif 'CP/EXECUTE/' in topic:
            return cls.Super.CP_EXECUTE
        elif 'CP/RESULT/SCHEDULE/' in topic:
            return cls.Super.CP_RESULT_SCHEDULE
        elif 'CP/RESULT/EXECUTE/' in topic:
            return cls.Super.CP_RESULT_EXECUTE
        elif 'CP/SERVICE_LIST/' in topic:
            return cls.Super.CP_SERVICE_LIST
        # EM
        if 'EM/REQUEST_CONNECTION/' in topic:
            return cls.WebClient.EM_REQUEST_CONNECTION
        elif 'EM/CONNECT_AS_OWNER/' in topic:
            return cls.WebClient.EM_CONNECT_AS_OWNER
        elif 'EM/SET_MEMBER_LEVEL/' in topic:
            return cls.WebClient.EM_SET_MEMBER_LEVEL
        elif 'EM/SET_MEMBER_STATE/' in topic:
            return cls.WebClient.EM_SET_MEMBER_STATE
        elif 'EM/DELETE_MEMBER/' in topic:
            return cls.WebClient.EM_DELETE_MEMBER
        elif 'EM/ADD_THING_FAVORITE/' in topic:
            return cls.WebClient.EM_ADD_THING_FAVORITE
        elif 'EM/DELETE_THING_FAVORITE/' in topic:
            return cls.WebClient.EM_DELETE_THING_FAVORITE
        elif 'EM/ADD_SERVICE_FAVORITE/' in topic:
            return cls.WebClient.EM_ADD_SERVICE_FAVORITE
        elif 'EM/DELETE_SERVICE_FAVORITE/' in topic:
            return cls.WebClient.EM_DELETE_SERVICE_FAVORITE
        elif 'EM/ADD_SCENARIO_FAVORITE/' in topic:
            return cls.WebClient.EM_ADD_SCENARIO_FAVORITE
        elif 'EM/DELETE_SCENARIO_FAVORITE/' in topic:
            return cls.WebClient.EM_DELETE_SCENARIO_FAVORITE
        elif 'EM/REGISTER_THING/' in topic:
            return cls.WebClient.EM_REGISTER_THING
        elif 'EM/UNREGISTER_THING/' in topic:
            return cls.WebClient.EM_UNREGISTER_THING
        elif 'EM/SET_NICK_NAME/' in topic:
            return cls.WebClient.EM_SET_NICK_NAME
        elif 'EM/SET_DESCRIPTION/' in topic:
            return cls.WebClient.EM_SET_DESCRIPTION
        elif 'EM/REQUEST_SCAN/' in topic:
            return cls.WebClient.EM_REQUEST_SCAN
        elif 'EM/OPEN_WINDOW_MATTER/' in topic:
            return cls.WebClient.EM_OPEN_WINDOW_MATTER
        elif 'EM/CLOSE_WINDOW_MATTER/' in topic:
            return cls.WebClient.EM_CLOSE_WINDOW_MATTER
        elif 'EM/EXECUTE/' in topic:
            return cls.WebClient.EM_INSTANT_EXECUTE
        elif 'EM/SET_ACCESS/' in topic:
            return cls.WebClient.EM_SET_ACCESS
        elif 'EM/SET_OPENED/' in topic:
            return cls.WebClient.EM_SET_OPENED
        elif 'EM/REFRESH/' in topic:
            return cls.WebClient.EM_REFRESH
        elif 'EM/GET_THING/' in topic:
            return cls.WebClient.EM_GET_THING
        elif 'EM/GET_SERVICE/' in topic:
            return cls.WebClient.EM_GET_SERVICE
        elif 'EM/GET_SCENARIO/' in topic:
            return cls.WebClient.EM_GET_SCENARIO
        elif 'EM/GET_MEMBER/' in topic:
            return cls.WebClient.EM_GET_MEMBER
        elif 'EM/GET_THING_LIST/' in topic:
            return cls.WebClient.EM_GET_THING_LIST
        elif 'EM/GET_SERVICE_LIST/' in topic:
            return cls.WebClient.EM_GET_SERVICE_LIST
        elif 'EM/GET_SCENARIO_LIST/' in topic:
            return cls.WebClient.EM_GET_SCENARIO_LIST
        elif 'EM/GET_MEMBER_LIST/' in topic:
            return cls.WebClient.EM_GET_MEMBER_LIST
        elif 'EM/SET_LOCATION/' in topic:
            return cls.WebClient.EM_SET_LOCATION
        elif 'EM/GET_LOCATION/' in topic:
            return cls.WebClient.EM_GET_LOCATION
        elif 'EM/GET_HOME/' in topic:
            return cls.WebClient.EM_GET_HOME
        elif 'EM/GET_FAVORITES/' in topic:
            return cls.WebClient.EM_GET_FAVORITES
        elif 'EM/RESET_HUB/' in topic:
            return cls.WebClient.EM_RESET_HUB
        # ME
        elif 'ME/RESULT/REQUEST_CONNECTION/' in topic:
            return cls.WebClient.ME_RESULT_REQUEST_CONNECTION
        elif 'ME/OWNER_BUTTON_CLICKED' in topic:
            return cls.WebClient.ME_OWNER_BUTTON_CLICKED
        elif 'ME/RESULT/CONNECT_AS_OWNER/' in topic:
            return cls.WebClient.ME_RESULT_CONNECT_AS_OWNER
        elif 'ME/RESULT/SET_MEMBER_LEVEL/' in topic:
            return cls.WebClient.ME_RESULT_SET_MEMBER_LEVEL
        elif 'ME/RESULT/SET_MEMBER_STATE/' in topic:
            return cls.WebClient.ME_RESULT_SET_MEMBER_STATE
        elif 'ME/RESULT/DELETE_MEMBER/' in topic:
            return cls.WebClient.ME_RESULT_DELETE_MEMBER
        elif 'ME/RESULT/ADD_THING_FAVORITE/' in topic:
            return cls.WebClient.ME_RESULT_ADD_THING_FAVORITE
        elif 'ME/RESULT/DELETE_THING_FAVORITE/' in topic:
            return cls.WebClient.ME_RESULT_DELETE_THING_FAVORITE
        elif 'ME/RESULT/ADD_SERVICE_FAVORITE/' in topic:
            return cls.WebClient.ME_RESULT_ADD_SERVICE_FAVORITE
        elif 'ME/RESULT/DELETE_SERVICE_FAVORITE/' in topic:
            return cls.WebClient.ME_RESULT_DELETE_SERVICE_FAVORITE
        elif 'ME/RESULT/ADD_SCENARIO_FAVORITE/' in topic:
            return cls.WebClient.ME_RESULT_ADD_SCENARIO_FAVORITE
        elif 'ME/RESULT/DELETE_SCENARIO_FAVORITE/' in topic:
            return cls.WebClient.ME_RESULT_DELETE_SCENARIO_FAVORITE
        elif 'ME/RESULT/REGISTER_THING/' in topic:
            return cls.WebClient.ME_RESULT_REGISTER_THING
        elif 'ME/RESULT/UNREGISTER_THING/' in topic:
            return cls.WebClient.ME_RESULT_UNREGISTER_THING
        elif 'ME/RESULT/SET_NICK_NAME/' in topic:
            return cls.WebClient.ME_RESULT_SET_NICK_NAME
        elif 'ME/RESULT/SET_DESCRIPTION/' in topic:
            return cls.WebClient.ME_RESULT_SET_DESCRIPTION
        elif 'ME/SCAN_RESULT/' in topic:
            return cls.WebClient.ME_SCAN_RESULT
        elif 'ME/RESULT/OPEN_WINDOW_MATTER/' in topic:
            return cls.WebClient.ME_RESULT_OPEN_WINDOW_MATTER
        elif 'ME/RESULT/CLOSE_WINDOW_MATTER/' in topic:
            return cls.WebClient.ME_RESULT_CLOSE_WINDOW_MATTER
        elif 'ME/RESULT/SET_ACCESS/' in topic:
            return cls.WebClient.ME_RESULT_SET_ACCESS
        elif 'ME/RESULT/SET_OPENED/' in topic:
            return cls.WebClient.ME_RESULT_SET_OPENED
        elif 'ME/RESULT/THING/' in topic:
            return cls.WebClient.ME_RESULT_THING
        elif 'ME/RESULT/SERVICE/' in topic:
            return cls.WebClient.ME_RESULT_SERVICE
        elif 'ME/RESULT/SCENARIO/' in topic:
            return cls.WebClient.ME_RESULT_SCENARIO
        elif 'ME/RESULT/MEMBER/' in topic:
            return cls.WebClient.ME_RESULT_MEMBER
        elif 'ME/RESULT/THING_LIST/' in topic:
            return cls.WebClient.ME_RESULT_THING_LIST
        elif 'ME/RESULT/SERVICE_LIST/' in topic:
            return cls.WebClient.ME_RESULT_SERVICE_LIST
        elif 'ME/RESULT/SCENARIO_LIST/' in topic:
            return cls.WebClient.ME_RESULT_SCENARIO_LIST
        elif 'ME/RESULT/MEMBER_LIST/' in topic:
            return cls.WebClient.ME_RESULT_MEMBER_LIST
        elif 'ME/RESULT/SET_LOCATION/' in topic:
            return cls.WebClient.ME_RESULT_SET_LOCATION
        elif 'ME/RESULT/GET_LOCATION/' in topic:
            return cls.WebClient.ME_RESULT_GET_LOCATION
        elif 'ME/RESULT/HOME/' in topic:
            return cls.WebClient.ME_RESULT_HOME
        elif 'ME/RESULT/GET_FAVORITES/' in topic:
            return cls.WebClient.ME_RESULT_GET_FAVORITES
        elif 'ME/NOTIFY_MESSAGE/' in topic:
            return cls.WebClient.ME_NOTIFY_MESSAGE
        elif 'ME/RESULT/RESET_HUB/' in topic:
            return cls.WebClient.ME_RESULT_RESET_HUB
        # TM/SN
        elif 'TM/SN/REGISTER/VALUE/' in topic:
            return cls.SN.TM_SN_REGISTER_VALUE
        elif 'TM/SN/REGISTER/VALUEDESC/' in topic:
            return cls.SN.TM_SN_REGISTER_VALUEDESC
        elif 'TM/SN/REGISTER/VALUETAG/' in topic:
            return cls.SN.TM_SN_REGISTER_VALUETAG
        elif 'TM/SN/REGISTER/FUNCTION/' in topic:
            return cls.SN.TM_SN_REGISTER_FUNCTION
        elif 'TM/SN/REGISTER/FUNCTIONDESC/' in topic:
            return cls.SN.TM_SN_REGISTER_FUNCTIONDESC
        elif 'TM/SN/REGISTER/FUNCTIONTAG/' in topic:
            return cls.SN.TM_SN_REGISTER_FUNCTIONTAG
        elif 'TM/SN/REGISTER/ARGUMENT/' in topic:
            return cls.SN.TM_SN_REGISTER_ARGUMENT
        elif 'TM/SN/REGISTER/ALIVECYCLE/' in topic:
            return cls.SN.TM_SN_REGISTER_ALIVECYCLE
        elif 'TM/SN/REGISTER/FINISH/' in topic:
            return cls.SN.TM_SN_REGISTER_FINISH
        else:
            return None


class MXRangeType(Enum):
    def _generate_next_value_(name: str, start, count, last_values):
        return name.lower()

    UNDEFINED = auto()
    ALL = auto()
    SINGLE = auto()

    @classmethod
    def get(cls, name: str) -> 'MXRangeType':
        try:
            return cls[name.upper()]
        except Exception:
            return cls.UNDEFINED


class HierarchyType(Enum):
    LOCAL = 0
    PARENT = 1
    CHILD = 2


class MXServiceType(Enum):
    def _generate_next_value_(name: str, start, count, last_values):
        return name.lower()

    UNDEFINED = auto()
    VALUE = auto()
    FUNCTION = auto()

    @classmethod
    def get(cls, name: str) -> 'MXServiceType':
        try:
            return cls[name.upper()]
        except Exception:
            return cls.UNDEFINED


class MXNetworkType(Enum):
    UNDEFINED = 0
    BLE = auto()
    BT = auto()
    MDNS = auto()
    ZIGBEE = auto()
    MATTER = auto()
    REST_API = auto()

    @classmethod
    def get(cls, name: str) -> 'MXNetworkType':
        try:
            return cls[name.upper()]
        except Exception:
            return cls.UNDEFINED


if __name__ == '__main__':
    a = BinaryBase64('thsvkd')
    b = BinaryBase64.from_normal_string('thsvkd')
    c = BinaryBase64.from_binary_string('dGhzdmtkdGhzdmtkdGhzdmtk')

    x = MXType.get(a)
    print(x)



import asyncio
from bless import BlessServer, BlessGATTCharacteristic, GATTCharacteristicProperties, GATTAttributePermissions

from big_thing_py.utils import *


class BLEErrorCode(Enum):
    NO_ERROR = 0
    FAIL = -1
    WIFI_PASSWORD_ERROR = -2
    WIFI_CONNECT_TIMEOUT = -3
    ALREADY_CONNECTED = -4
    WIFI_NOT_FOUND = -5
    WIFI_CREDENTIAL_NOT_SET = -6
    BROKER_NOT_SET = -7


class Characteristic:
    def __init__(self, uuid: str, properties: GATTCharacteristicProperties, permissions: GATTAttributePermissions, value: bytearray = None):
        self.uuid = uuid
        self.properties = properties
        self.permissions = permissions
        self.value = value


class Service:
    def __init__(self, uuid: str, characteristics: List[Characteristic]):
        self.uuid = uuid
        self.characteristics = characteristics


class DeviceWifiService(Service):
    UUID = '640F0000-0000-0000-0000-000000000000'

    class SetWifiSSIDCharacteristic(Characteristic):
        def __init__(self):
            super().__init__(
                uuid='640F0001-0000-0000-0000-000000000000',
                properties=GATTCharacteristicProperties.write,
                permissions=GATTAttributePermissions.writeable,
            )

    class SetWifiPWCharacteristic(Characteristic):
        def __init__(self):
            super().__init__(
                uuid='640F0002-0000-0000-0000-000000000000',
                properties=GATTCharacteristicProperties.write,
                permissions=GATTAttributePermissions.writeable,
            )

    class SetBrokerInfoCharacteristic(Characteristic):
        def __init__(self):
            super().__init__(
                uuid='640F0003-0000-0000-0000-000000000000',
                properties=GATTCharacteristicProperties.write,
                permissions=GATTAttributePermissions.writeable,
            )

    class ConnectWifiCharacteristic(Characteristic):
        def __init__(self):
            super().__init__(
                uuid='640F0004-0000-0000-0000-000000000000',
                properties=GATTCharacteristicProperties.write,
                permissions=GATTAttributePermissions.writeable,
            )

    class ThingIDCharacteristic(Characteristic):
        def __init__(self):
            super().__init__(
                uuid='640F0005-0000-0000-0000-000000000000',
                properties=GATTCharacteristicProperties.read,
                permissions=GATTAttributePermissions.readable,
            )

    class ErrorCodeCharacteristic(Characteristic):
        def __init__(self):
            super().__init__(
                uuid='640F0006-0000-0000-0000-000000000000',
                properties=GATTCharacteristicProperties.read,
                permissions=GATTAttributePermissions.readable,
            )

    def __init__(self):
        characteristics = [
            self.SetWifiSSIDCharacteristic(),
            self.SetWifiPWCharacteristic(),
            self.SetBrokerInfoCharacteristic(),
            self.ConnectWifiCharacteristic(),
            # self.ThingIDCharacteristic(),
            self.ErrorCodeCharacteristic(),
        ]
        super().__init__(DeviceWifiService.UUID, characteristics)


class BLEAdvertiser:
    def __init__(self, server_name: str) -> None:
        self._server_name = server_name
        self._server: BlessServer = None
        self._trigger = asyncio.Event()

    def _read_request(self, characteristic: BlessGATTCharacteristic, **kwargs) -> bytearray:
        MXLOG_DEBUG(f'Reading {characteristic.value}')
        return characteristic.value

    def _write_request(self, characteristic: BlessGATTCharacteristic, value: Any, **kwargs):
        try:
            characteristic.value = value
            uuid = characteristic.uuid.upper()
            if uuid == DeviceWifiService.SetWifiSSIDCharacteristic().uuid:
                MXLOG_DEBUG(f'WiFi SSID set: {self._server.get_characteristic(DeviceWifiService.SetWifiSSIDCharacteristic().uuid).value}')
            elif uuid == DeviceWifiService.SetWifiPWCharacteristic().uuid:
                MXLOG_DEBUG(f'WiFi PW set: {self._server.get_characteristic(DeviceWifiService.SetWifiPWCharacteristic().uuid).value}')
            elif uuid == DeviceWifiService.SetBrokerInfoCharacteristic().uuid:
                MXLOG_DEBUG(f'Broker Info set: {self._server.get_characteristic(DeviceWifiService.SetBrokerInfoCharacteristic().uuid).value}')
            elif uuid == DeviceWifiService.ConnectWifiCharacteristic().uuid:
                ssid = self._server.get_characteristic(DeviceWifiService.SetWifiSSIDCharacteristic().uuid).value
                pw = self._server.get_characteristic(DeviceWifiService.SetWifiPWCharacteristic().uuid).value
                broker = self._server.get_characteristic(DeviceWifiService.SetBrokerInfoCharacteristic().uuid).value
                if ssid is None or pw is None:
                    MXLOG_DEBUG(f'WiFi credentials not set... ssid: {ssid}, pw: {pw}')
                    self._server.update_value(
                        DeviceWifiService.ErrorCodeCharacteristic().uuid, BLEErrorCode.WIFI_CREDENTIAL_NOT_SET.value.to_bytes(2, 'little')
                    )
                    return
                elif broker is None:
                    MXLOG_DEBUG(f'Broker info not set... broker: {broker}')
                    self._server.update_value(
                        DeviceWifiService.ErrorCodeCharacteristic().uuid, BLEErrorCode.BROKER_NOT_SET.value.to_bytes(2, 'little')
                    )
                    return
                else:
                    MXLOG_DEBUG('wifi credentials & broker info is set!', 'green')
                    self._trigger.set()
        except Exception as e:
            MXLOG_DEBUG(f'Error occurred while writing characteristic: {e}', 'red')
            self._server.update_value(DeviceWifiService.ErrorCodeCharacteristic().uuid, BLEErrorCode.FAIL.value.to_bytes(2, 'little'))

    async def _add_service(self, service: Service):
        await self._server.add_new_service(service.uuid)
        for char in service.characteristics:
            await self._server.add_new_characteristic(service.uuid, char.uuid, char.properties, char.value, char.permissions)

    async def _add_thing_id_characteristic(self, thing_id: str):
        await self._server.add_new_characteristic(
            service_uuid=DeviceWifiService.UUID,
            char_uuid=DeviceWifiService.ThingIDCharacteristic().uuid,
            properties=GATTCharacteristicProperties.read,
            value=bytearray(thing_id.encode()),
            permissions=GATTAttributePermissions.readable,
        )

    async def start(self, thing_id: str):
        MXLOG_DEBUG('Starting BLE advertiser...')
        self._trigger.clear()
        self._server = BlessServer(name=self._server_name)
        self._server.read_request_func = self._read_request
        self._server.write_request_func = self._write_request

        await self._add_service(DeviceWifiService())
        await self._add_thing_id_characteristic(thing_id)

        await self._server.start()
        MXLOG_DEBUG('BLE Advertising started...')

    async def is_advertising(self) -> bool:
        if self._server is None:
            return False

        return await self._server.is_advertising()

    async def wait_until_wifi_credentials_set(self, timeout: float = 30) -> Tuple[str, str, str, BLEErrorCode]:

        async def wrapper() -> Tuple[str, str, str, BLEErrorCode]:
            await self._trigger.wait()
            ssid = self._server.get_characteristic(DeviceWifiService.SetWifiSSIDCharacteristic().uuid).value.decode()
            pw = self._server.get_characteristic(DeviceWifiService.SetWifiPWCharacteristic().uuid).value.decode()
            broker = self._server.get_characteristic(DeviceWifiService.SetBrokerInfoCharacteristic().uuid).value.decode()
            error_code = BLEErrorCode(
                int.from_bytes(self._server.get_characteristic(DeviceWifiService.ErrorCodeCharacteristic().uuid).value, 'little')
            )
            return (ssid, pw, broker, error_code)

        try:
            ssid, pw, broker, error_code = await asyncio.wait_for(wrapper(), timeout)
            return (ssid, pw, broker, error_code)
        except asyncio.TimeoutError:
            return ('', '', '', BLEErrorCode.WIFI_CONNECT_TIMEOUT)

    async def stop(self):
        await self._server.stop()
        MXLOG_DEBUG('BLE Advertising stopped...')

    async def is_connected(self):
        return await self._server.is_connected()


if __name__ == '__main__':
    START_LOGGER()

    async def run():
        ble_advertiser = BLEAdvertiser(server_name=f'Joi Smart Device {get_mac_address()}')
        await ble_advertiser.start()
        ssid, pw, broker, error_code = await ble_advertiser.wait_until_wifi_credentials_set()
        MXLOG_DEBUG(f'WiFi credentials set: ssid: {ssid}, pw: {pw}, broker: {broker}, error_code: {error_code}')
        await ble_advertiser.stop()

    async def test_smart_device(server_name: str = f'Joi Smart Device {get_mac_address()}'):
        '''
        NOTE: this test function should be run on a separate device
        '''
        from bleak import BleakClient, BleakScanner

        device_address = None

        devices = await BleakScanner.discover()
        for device in devices:
            if device.name == server_name:
                print(f'Found BLE server! name: {device.name}, address: {device.address}')
                device_address = device.address
                break
        else:
            print(f"Cannot find {server_name}")
            return

        ssid_characteristic_uuid = DeviceWifiService.SetWifiSSIDCharacteristic().uuid
        pw_characteristic_uuid = DeviceWifiService.SetWifiPWCharacteristic().uuid
        broker_characteristic_uuid = DeviceWifiService.SetBrokerInfoCharacteristic().uuid
        connect_wifi_characteristic_uuid = DeviceWifiService.ConnectWifiCharacteristic().uuid

        ssid_value = b"MySmaX-office5G"
        pw_value = b"/PeaCE/#1"
        broker_value = b"192.168.0.75:8883"

        async with BleakClient(device_address) as client:
            if client.is_connected:
                print(f"Connected to {device_address}")

                await client.write_gatt_char(ssid_characteristic_uuid, ssid_value)
                print("WiFi SSID set")

                await client.write_gatt_char(pw_characteristic_uuid, pw_value)
                print("WiFi password set")

                await client.write_gatt_char(broker_characteristic_uuid, broker_value)
                print("Broker info set")

                await client.write_gatt_char(connect_wifi_characteristic_uuid, bytearray([0x00]))
                print("WiFi connection attempt")

    asyncio.run(test_smart_device())



from big_thing_py.common import *

import inspect
import socket
import json
import time
import platform
import os
import copy
import random
from pathlib import Path
import getmac
import re
import sys
import typing
import importlib.metadata
import functools
from termcolor import cprint


# def static_vars(**kwargs):
#     def decorate(func):
#         for k in kwargs:
#             setattr(func, k, kwargs[k])
#         return func

#     return decorate


def static_vars(**kwargs):
    def decorator(func):
        for k, v in kwargs.items():
            setattr(func, k, v)

        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return decorator


def get_type_name(typ: typing.Type) -> str:
    origin = typing.get_origin(typ)
    if origin is None:
        return typ.__name__
    args = typing.get_args(typ)
    args_str = ", ".join(get_type_name(arg) for arg in args)
    return f"{origin.__name__}[{args_str}]"


def get_function_info(func: Callable) -> Dict[str, Union[str, List[Tuple[str, Any]]]]:
    sig = inspect.signature(func)
    params = [(param.name, param.annotation) for param in sig.parameters.values()]
    ret_type = sig.return_annotation
    if not ret_type is None:
        if isinstance(ret_type, str):
            try:
                ret_type = eval(ret_type)
            except NameError:
                pass
        ret_type = get_type_name(ret_type)
    return dict(name=func.__name__, args=params, return_type=ret_type)


def get_current_function_name() -> str:
    return inspect.stack()[1].function


def get_upper_function_name(step: int = 1, frame=None) -> str:
    if frame is None:
        frame = inspect.stack()[1].frame
    if step == 1:
        co_name = frame.f_back.f_code.co_name
    elif step > 1:
        co_name = get_upper_function_name(step - 1, frame.f_back)
    else:
        co_name = 'too many steps... return MAX upper function name'
    return co_name


def get_function_return_type(func: Callable, to_mxtype: bool = True) -> Union[MXType, str]:
    return_type = inspect.signature(func).return_annotation
    if return_type is inspect.Signature.empty:
        raise MXNotSupportedError('Return type is not defined!!!')

    if to_mxtype:
        return MXType.get(return_type)
    else:
        return return_type


def get_mac_address(interface: str = None) -> str:
    mac_address = getmac.get_mac_address(interface)

    if mac_address:
        mac_address = mac_address.replace(':', '').upper()
        return mac_address
    else:
        return None


def transform_mac_address(mac_address: str, to_colon: bool = True) -> str:
    if to_colon:
        return ':'.join(mac_address[i : i + 2] for i in range(0, len(mac_address), 2))
    else:
        return mac_address.replace(':', '')


def get_current_datetime(mode: TimeFormat = TimeFormat.UNIXTIME) -> Union[str, float]:
    if mode == TimeFormat.DATETIME1:
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    elif mode == TimeFormat.DATETIME2:
        return time.strftime("%Y%m%d_%H%M%S", time.localtime())
    elif mode == TimeFormat.DATE:
        return time.strftime("%Y-%m-%d", time.localtime())
    elif mode == TimeFormat.TIME:
        return time.strftime("%H:%M:%S", time.localtime())
    elif mode == TimeFormat.UNIXTIME:
        return time.time()
    return time.time()


def type_converter(in_type: Union[MXType, type, str, None]) -> Union[type, MXType]:
    if type(in_type) == MXType:
        if in_type in (MXType.BINARY, MXType.STRING):
            return str
        elif in_type == MXType.BOOL:
            return bool
        elif in_type == MXType.DOUBLE:
            return float
        elif in_type == MXType.INTEGER:
            return int
        elif in_type in (MXType.VOID, MXType.UNDEFINED):
            return None
        else:
            raise MXNotSupportedError('Unexpected python type!!!')
    elif type(in_type) == type:
        if in_type == int:
            return MXType.INTEGER
        elif in_type == float:
            return MXType.DOUBLE
        elif in_type == bool:
            return MXType.BOOL
        elif in_type == bytes:
            return MXType.BINARY
        elif in_type == str:
            return MXType.STRING
        elif in_type == type(None):
            return MXType.VOID
        else:
            raise MXNotSupportedError('Unexpected MXType type!!!')
    elif type(in_type) == str:
        if in_type == 'int':
            return MXType.INTEGER
        elif in_type == 'void':
            return MXType.VOID
        elif in_type == 'double':
            return MXType.DOUBLE
        elif in_type == 'bool':
            return MXType.BOOL
        elif in_type == 'binary':
            return MXType.BINARY
        elif in_type == 'string':
            return MXType.STRING
        elif in_type == 'undefined':
            return MXType.UNDEFINED
        else:
            raise MXNotSupportedError('Unexpected MXType type!!!')
    elif in_type == None or type(in_type) == type(None):
        return MXType.VOID


def json_file_read(path: str, mode: str = 'r') -> Union[dict, bool]:
    try:
        with open(path, mode) as f:
            return json.load(f)
    except FileNotFoundError:
        return False


def json_file_write(path: str, data: Union[str, dict], mode: str = 'w', indent: int = 4) -> None:
    with open(path, mode) as f:
        if isinstance(data, (dict, str)):
            json.dump(data, f, indent=indent)
        else:
            raise Exception(f'common_util.json_file_write: data type error - {type(data)}')


def get_project_root(project_name: str = 'big-thing-py') -> Path:
    start_path = Path(__file__)
    while True:
        if str(start_path).split('/')[-1] == project_name:
            return str(start_path)
        else:
            start_path = start_path.parent


def check_valid_identifier(identifier: str) -> MXErrorCode:
    pattern = r'^[_a-zA-Z가-힣][_a-zA-Z0-9가-힣]*$'

    if len(identifier) >= 256:
        return MXErrorCode.TOO_LONG_IDENTIFIER
    elif not bool(re.match(pattern, identifier)):
        return MXErrorCode.INVALID_DATA
    else:
        return MXErrorCode.NO_ERROR


def convert_special_char(char: str) -> str:
    # conversion_dict = {
    #     ' ': 'space',
    #     '!': 'exclamation_mark',
    #     '"': 'double_quote',
    #     '#': 'hash',
    #     '$': 'dollar',
    #     '%': 'percent',
    #     '&': 'ampersand',
    #     "'": 'single_quote',
    #     '(': 'left_parenthesis',
    #     ')': 'right_parenthesis',
    #     '*': 'asterisk',
    #     '+': 'plus',
    #     ',': 'comma',
    #     '-': 'hyphen',
    #     '.': 'period',
    #     '/': 'slash',
    #     ':': 'colon',
    #     ';': 'semicolon',
    #     '<': 'less_than',
    #     '=': 'equals',
    #     '>': 'greater_than',
    #     '?': 'question_mark',
    #     '@': 'at_sign',
    #     '[': 'left_bracket',
    #     '\\': 'backslash',
    #     ']': 'right_bracket',
    #     '^': 'caret',
    #     '`': 'backtick',
    #     '{': 'left_brace',
    #     '|': 'vertical_bar',
    #     '}': 'right_brace',
    #     '~': 'tilde'
    # }
    conversion_dict = {
        ' ': 'sp',
        '!': 'ex',
        '"': 'dq',
        '#': 'hs',
        '$': 'dl',
        '%': 'pc',
        '&': 'am',
        "'": 'sq',
        '(': 'lp',
        ')': 'rp',
        '*': 'as',
        '+': 'pl',
        ',': 'cm',
        '-': 'mn',
        '.': 'pd',
        '/': 'sl',
        ':': 'cl',
        ';': 'sc',
        '<': 'lt',
        '=': 'eq',
        '>': 'gt',
        '?': 'qm',
        '@': 'at',
        '[': 'lb',
        '\\': 'bs',
        ']': 'rb',
        '^': 'cr',
        '`': 'bt',
        '{': 'lc',
        '|': 'vb',
        '}': 'rc',
        '~': 'td',
    }
    if char in conversion_dict:
        return f'_{conversion_dict[char]}_'
    else:
        return char


def convert_to_valid_string(input_string: str) -> str:
    converted_chars = [convert_special_char(char) for char in input_string]
    return ''.join(converted_chars)


def convert_url_to_ip(url: str) -> str:
    ip_list = url.split(".")
    not_ip = False

    for ip_token in ip_list:
        if not ip_token.isdigit():
            not_ip = True
            break

    if not_ip:
        host_ip = socket.gethostbyname(url.strip())
        return host_ip
    elif not all([(0 <= int(ip_token) <= 255) for ip_token in ip_list]) or len(ip_list) != 4:
        raise Exception('wrong ip format')
    else:
        return url


def get_local_ip() -> str:
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.connect(('192.168.0.1', 80))
            local_ip = s.getsockname()[0]
            return local_ip
    except Exception as e:
        return f"Error occurred: {e}"


def check_python_version() -> bool:
    if not (sys.version_info[0] >= 3 and sys.version_info[1] > 6):
        return False

    return True


def is_valid_ip_address(ip_address: str) -> bool:
    if ip_address:
        octets = ip_address.split('.')
    else:
        return False

    if len(octets) != 4:
        return False
    for octet in octets:
        try:
            value = int(octet)
        except ValueError:
            return False
        if not 0 <= value <= 255:
            return False
    return True


def is_raspberry_pi():
    try:
        with open('/proc/device-tree/model', 'r') as f:
            return 'raspberry pi' in f.read().lower()
    except FileNotFoundError:
        return False


def check_os_architecture() -> str:
    return platform.architecture()[0]


def install_missing_package(error):
    import subprocess

    package_name = str(error).split("'")[1]
    subprocess.run([sys.executable, "-m", "pip", "install", package_name])


def dummy_func(arg_list: list):
    args = ', '.join([str(arg) for arg in arg_list])
    lambda_str = f"lambda {args}: None"
    exec(lambda_str)
    return eval(lambda_str)


def sdk_version(package_name='big-thing-py'):
    try:
        return importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        return None


def prune_payload(payload: str, mode: MXPrintMode = MXPrintMode.ABBR) -> str:
    if mode == MXPrintMode.SKIP:
        payload = colored(f'skip... (mode={mode})', 'yellow')
    elif mode == MXPrintMode.ABBR:
        if payload.count('\n') > 10:
            payload = '\n'.join(payload.split('\n')[:10]) + '\n' + colored(f'skip... (mode={mode})', 'yellow')
        elif len(payload) > 1000:
            payload = payload[:1000] + '\n' + colored(f'skip... (mode={mode})', 'yellow')
        else:
            pass
    elif mode == MXPrintMode.FULL:
        pass
    else:
        raise Exception(f'[prune_payload] Unknown mode!!! mode should be [skip|abbr|full] mode : {mode}', 'red')

    return payload


def print_func_info(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        cprint(f'[print_args] func: {func.__name__}', 'cyan')
        cprint(f'[print_args] args: {args}', 'cyan')
        cprint(f'[print_args] kwargs: {kwargs}', 'cyan')
        return func(*args, **kwargs)

    return wrapper


if __name__ == '__main__':
    pass
